#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Interpolation method based on:
'Hex-Splines: A Novel Spline Family for Hexagonal Lattices'
van de Ville et al. IEEE Transactions on Image Processing 2004, 13, 6
"""

from __future__ import print_function

import math

import numpy as np

# Normalized hexagon geometry
M_SQRT3 = math.sqrt(3)
H_HEX = 0.5
R_HEX = 1 / M_SQRT3
A_HEX = 0.5 * H_HEX * R_HEX
HA_HEX = 6 * A_HEX  # detR0 == ha


def calc_matrix(nrow, ncol, grid_type=2):
    """

    Parameters
    ----------
    nrow : int
    ncol : int
    grid_type : int

    Returns
    -------

    """

    rr0 = np.array([[M_SQRT3 / 2, 0], [-0.5, 1]])  # Unit scale

    if grid_type == 2:
        f = 0
    else:
        f = 1

    kcol = []
    krow = []
    for i in range(ncol):
        s = (i + f * (i % 2)) // 2
        for j in range(nrow):
            kcol.append(i)
            krow.append(j+s)

    sl = np.array([kcol, krow])  # x y
    r0l = np.dot(rr0, sl)
    # r0l = R0 @ sl
    return r0l


def calc_lcb_grid(scale=1.0):
    """

    Parameters
    ----------
    scale : float

    Returns
    -------

    """

    grid_type = 2  # Values for MEGARA
    ncol = 27  #
    nrow = 21  #
    r0l = calc_matrix(nrow, ncol, grid_type=grid_type)
    # r0l = R0 @ sl
    spos_x = scale * (r0l[0] - r0l[0].max() / 2)
    spos_y = scale * (r0l[1] - r0l[1].max() / 2)

    return spos_x, spos_y


def hexgrid_extremes(r0l, target_scale):
    """

    Parameters
    ----------
    r0l
    target_scale : float

    Returns
    -------

    """
    # geometry
    # ha_hex = 6 * a_hex # detR0 == ha
    # compute extremes of hexgrid to rectangular grid
    # with pixel size 'scale'
    x0min, y0min = r0l.min(axis=1)
    x0max, y0max = r0l.max(axis=1)
    y1min = y0min - H_HEX
    y1max = y0max + H_HEX
    x1min = x0min - R_HEX
    x1max = x0max + R_HEX

    j1min = int(math.floor(x1min / target_scale + 0.5))
    i1min = int(math.floor(y1min / target_scale + 0.5))
    j1max = int(math.ceil(x1max / target_scale - 0.5))
    i1max = int(math.ceil(y1max / target_scale - 0.5))
    return (i1min, i1max), (j1min, j1max)


def connected6(i, j):
    """Connected pixels in a hexagonal grid"""
    for a, b in [(0, 1), (0, -1), (1, 0), (-1, 0), (1, -1), (-1, 1)]:
        u, v = i + a, j + b
        yield u, v


def neigh6(i, j, level=1):
    """neighborhood in 6-connection"""
    # square
    for m in range(0, level + 1):
        for n in range(0, level + 1):
            yield i - m, j + n
    for k in range(level):
        for ll in range(1, level + 1):
            yield i + 1 + k, j - 1 - k + ll
            yield i + 1 + k - ll, j - 1 - k
        yield i + 1 + k, j - 1 - k




from ..detector import MegaraDetector, ReadParams


def create_detector():
    DSHAPE = (2056 * 2, 2048 * 2)
    PSCAN = 50
    OSCAN = 50
    qe = 1.0
    dcurrent = 3.0 / 3600

    readpars1 = ReadParams(gain=1.0, ron=2.0, bias=1000.0)
    readpars2 = ReadParams(gain=1.0, ron=2.0, bias=1005.0)

    detector = MegaraDetector(
        'Detector',
        DSHAPE, OSCAN, PSCAN, qe=qe, dark=dcurrent,
        readpars1=readpars1, readpars2=readpars2, bins='11'
    )
    return detector


def test_detector_shape():
    det = create_detector()
    img = det.readout()
    assert img.shape == (4212, 4196)



import pytest
import numpy as np
import astropy.wcs

from megaradrp.tests.simpleobj import create_spec_header2, create_sky_header2
from megaradrp.processing.wavecalibration import header_add_barycentric_correction

from ..cube import create_cube, merge_wcs


def test_create_cube_raise():
    with pytest.raises(ValueError):
        create_cube(None, None, 3)


def test_sub_wcs():
    hdr_sky = create_sky_header2()
    hdr_spec = create_spec_header2()
    hdr_spec = header_add_barycentric_correction(hdr_spec)
    wcs_sky = astropy.wcs.WCS(header=hdr_sky)
    wcs_spec = astropy.wcs.WCS(header=hdr_spec, key='B')
    wcs3 = wcs_sky.sub([1,2,0])
    wcs3.wcs.ctype[2] = wcs_spec.wcs.ctype[0]
    wcs3.wcs.crval[2] = wcs_spec.wcs.crval[0]
    wcs3.wcs.crpix[2] = wcs_spec.wcs.crpix[0]
    wcs3.wcs.cdelt[2] = wcs_spec.wcs.cdelt[0]
    wcs3.wcs.cunit[2] = wcs_spec.wcs.cunit[0]
    wcs3.wcs.specsys = wcs_spec.wcs.specsys
    wcs3.wcs.ssysobs = wcs_spec.wcs.ssysobs
    wcs3.wcs.velosys = wcs_spec.wcs.velosys
    hdr3 = wcs3.to_header(key='B')
    assert True


def test_merge_wcs():
    hdr_spec = create_spec_header2()
    hdr_spec = header_add_barycentric_correction(hdr_spec)
    hdr_sky = create_sky_header2()
    out = hdr_spec.copy()
    merge_wcs(hdr_sky, hdr_spec, out=out)

    wcs0 = astropy.wcs.WCS(header=out)
    wcsB = astropy.wcs.WCS(header=out, key='B')

    assert list(wcs0.wcs.ctype) == ['RA---TAN', 'DEC--TAN', 'AWAV']
    assert np.allclose([out['CRVAL1'], out['CRVAL2'], out['CRVAL3']],
                       [hdr_sky['CRVAL1'], hdr_sky['CRVAL2'], hdr_spec['CRVAL1']])
    assert np.allclose([out['CDELT1'], out['CDELT2'], out['CDELT3']],
                       [hdr_sky['CDELT1'], hdr_sky['CDELT2'], hdr_spec['CDELT1']])
    assert list(wcsB.wcs.ctype) == ['RA---TAN', 'DEC--TAN', 'AWAV']
    assert np.allclose([out['CRVAL1B'], out['CRVAL2B'], out['CRVAL3B']],
                       [hdr_sky['CRVAL1'], hdr_sky['CRVAL2'], hdr_spec['CRVAL1B']])
    assert np.allclose([out['CDELT1B'], out['CDELT2B'], out['CDELT3B']],
                       [hdr_sky['CDELT1'], hdr_sky['CDELT2'], hdr_spec['CDELT1B']])



#
# Copyright 2011-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""MOS Standard Star Image Recipe for Megara"""


import astropy.io.fits as fits
import astropy.units as u
import astropy.wcs
from astropy import constants as const
from scipy.interpolate import interp1d

from numina.array.numsplines import AdaptiveLSQUnivariateSpline
from numina.core import Result, Parameter
from numina.core.requirements import Requirement
from numina.exceptions import RecipeError
from numina.types.array import ArrayType

from megaradrp.instrument.focalplane import FocalPlaneConf
from megaradrp.ntypes import Point2D
from megaradrp.ntypes import ProcessedRSS, ProcessedFrame, ProcessedSpectrum
from megaradrp.ntypes import ReferenceSpectrumTable, ReferenceExtinctionTable
from megaradrp.ntypes import MasterSensitivity
from megaradrp.processing.extractobj import extract_star, generate_sensitivity
from megaradrp.processing.extractobj import mix_values, compute_broadening
from megaradrp.processing.centroid import calc_centroid_brightest
from megaradrp.recipes.scientific.base import ImageRecipe


class MOSStandardRecipe(ImageRecipe):
    """Process MOS Standard Star Recipe.

    This recipe processes a set of images
    obtained in **MOS Stardard Star image** mode and returns
    the total flux of the star.

    See Also
    --------
    megaradrp.recipes.calibration.lcbstdstar.LCBStandardRecipe

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are then stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_apertures` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned ini the
    field `final_rss` of the recipe result.

    The flux of the star is computed by adding the 7 fibers corresponding to the bundle
    containing the star and returned as `star_spectrum`.

    """
    position = Requirement(Point2D, "Position of the reference object", optional=True)
    # nrings = 1
    reference_spectrum = Requirement(ReferenceSpectrumTable, "Spectrum of reference star")
    reference_spectrum_velocity = Parameter(0.0, 'Radial velocity of reference spectrum')
    reference_extinction = Requirement(ReferenceExtinctionTable, "Reference extinction")
    degrade_resolution_target = Parameter('object', 'Spectrum with higher resolution',
                                          choices=['object']
                                          )
    # TODO: Implement the possibility of the reference having higher resolution
    # degrade_resolution_target = Parameter('object', 'Spectrum with higher resolution',
    #                                       choices=['object', 'reference']
    #                                      )
    degrade_resolution_method = Parameter('fixed', 'Method to degrade the resolution',
                                          choices=['none', 'fixed', 'auto']
                                          )
    sigma_resolution = Parameter(20.0, 'sigma Gaussian filter to degrade resolution ')

    reduced_image = Result(ProcessedFrame)
    final_rss = Result(ProcessedRSS)
    reduced_rss = Result(ProcessedRSS)
    sky_rss = Result(ProcessedRSS)
    star_spectrum = Result(ProcessedSpectrum)
    master_sensitivity = Result(MasterSensitivity)
    sensitivity_raw = Result(ProcessedSpectrum)
    fiber_ids = Result(ArrayType)
    sigma = Result(float)

    def run(self, rinput):

        self.logger.info('starting MOSStandardRecipe reduction')

        # Try to guard against receiving here something
        # that is not in magAB
        # TODO: implement this in ReferenceSpectrumTable
        maxm = max(rinput.reference_spectrum[:, 1])
        if maxm > 100:
            # If the column here has values greater than 100
            # this could not be a magnitude
            raise RecipeError("the maximum flux of 'reference_spectrum' is > 100, "
                              "check the flux unit (it has to be magAB)")

        reduced2d, rss_data = super(MOSStandardRecipe, self).base_run(rinput)

        self.logger.info('start sky subtraction')
        final, origin, sky = self.run_sky_subtraction(
            rss_data,
            sky_rss=rinput.sky_rss,
            ignored_sky_bundles=rinput.ignored_sky_bundles
        )
        self.logger.info('end sky subtraction')

        # 1 + 6  for first ring
        # 1 + 6  + 12  for second ring
        # 1 + 6  + 12  + 18 for third ring
        # 1 + 6 * Sum_i=0^n =  1 + 3 * n * (n +1)
        # In MOS, only 1 ring around central point
        # If position is None, find the brightest spaxel
        # and use the centroid
        if rinput.position is None:
            self.logger.info('finding centroid of brightest spaxel')
            extraction_region = [1000, 3000]
            nrings = rinput.nrings
            position = calc_centroid_brightest(final, extraction_region, nrings)
        else:
            position = rinput.position
        self.logger.info('central position is %s', position)

        self.logger.debug('adding %d nrings', 1)
        npoints = 7
        self.logger.debug('adding %d fibers', npoints)

        fp_conf = FocalPlaneConf.from_img(final)
        spectra_pack = extract_star(final, position, npoints,
                                    fp_conf, logger=self.logger)

        spectrum, colids, wl_cover1, wl_cover2 = spectra_pack
        star_spectrum = fits.PrimaryHDU(spectrum, header=final[0].header)

        rad_vel = rinput.reference_spectrum_velocity * u.km / u.s
        factor = 1 + rad_vel / const.c

        star_interp = interp1d(rinput.reference_spectrum[:, 0] / factor,
                               rinput.reference_spectrum[:, 1])
        
        extinc_interp = interp1d(rinput.reference_extinction[:, 0],
                                 rinput.reference_extinction[:, 1])

        fiber_ids = [colid + 1 for colid in colids]

        wcsl = astropy.wcs.WCS(final[0].header)
        wl_aa, response_m, response_r = mix_values(wcsl, spectrum, star_interp)
        if rinput.degrade_resolution_method == 'none':
            sigma = 0
            self.logger.info('no broadening')
        elif rinput.degrade_resolution_method == 'fixed':
            sigma = rinput.sigma_resolution
            self.logger.info('fixed sigma=%3.0f', sigma)
        elif rinput.degrade_resolution_method == 'auto':
            self.logger.info('compute auto broadening')
            offset_broad, sigma_broad = compute_broadening(
                response_r.copy(), response_m.copy(), sigmalist=range(1, 101),
                remove_mean=False, frac_cosbell=0.10, zero_padding=50,
                fminmax=(0.003, 0.3), naround_zero=25, nfit_peak=21
            )
            sigma = sigma_broad
            self.logger.info('computed sigma=%3.0f', sigma)
        else:
            msg = f"'degrade_resolution_method' has value {rinput.degrade_resolution_method}"
            raise ValueError(msg)

        sens_raw = generate_sensitivity(final, spectrum, star_interp, extinc_interp, wl_cover1, wl_cover2, sigma)

        # Compute smoothed version
        self.logger.info('compute smoothed sensitivity')

        sens = sens_raw.copy()
        i_knots = 3
        self.logger.debug('using sdaptive spline with t=%d interior knots', i_knots)
        spl = AdaptiveLSQUnivariateSpline(x=wl_aa.value, y=sens_raw.data, t=i_knots)
        sens.data = spl(wl_aa.value)

        if self.intermediate_results:
            import matplotlib.pyplot as plt
            plt.plot(wl_aa, sens_raw.data, 'b')
            plt.plot(wl_aa, sens.data, 'r')
            plt.savefig('smoothed.png')
            plt.close()

        self.logger.info('end MOSStandardRecipe reduction')

        return self.create_result(
            reduced_image=reduced2d,
            final_rss=final,
            reduced_rss=origin,
            sky_rss=sky,
            star_spectrum=star_spectrum,
            master_sensitivity=sens,
            sensitivity_raw=sens_raw,
            fiber_ids=fiber_ids,
            sigma=sigma
        )



#
# Copyright 2016-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Validators for Observing modes"""


import pkgutil
from io import StringIO

import json
import jsonschema

from numina.exceptions import ValidationError

from megaradrp.datatype import MegaraDataType


def validate_focus(mode, obresult):
    """Validate FOCUS_SPECTROGRAPH"""
    image_groups = {}
    for idx, frame in enumerate(obresult.frames):
        with frame.open() as img:
            try:
                focus_val = img[0].header['FOCUS']
                # FIXME: This should be done using an scheme for MEGARA
                if not isinstance(focus_val, (int, float)):
                    raise ValidationError(f"FOCUS must be integer, not {type(focus_val)}")
                if focus_val not in image_groups:
                    image_groups[focus_val] = []
                image_groups[focus_val].append(frame)
            except Exception as exc:
                raise ValidationError from exc

    if len(image_groups) < 2:
        raise ValidationError(f'We have only {len(image_groups)} different focus in OB')

    return True


def validate_key(mode, obresult, key):
    """Validate key"""

    # Assume that the individual images are valid IMG_COMP
    # check consistency of key
    kval = []
    for idx, frame in enumerate(obresult.frames):
        # SPECLAMP values
        with frame.open() as img:
            try:
                spec_val = img[0].header[key]
                kval.append(spec_val)
            except Exception as exc:
                raise ValidationError from exc

    if kval[:-1] == kval[1:]:
        return True
    else:
        raise ValidationError(f"{key} value is incorrect")


def validate_arc(mode, obresult):
    """Validate ARC_CALIBRATION"""

    # Assume that the individual images are valid IMG_COMP
    return validate_key(mode, obresult, 'SPECLAMP')


def validate_flat(mode, obresult):
    """Validate FLAT"""

    # Assume that the individual images are valid IMG_COMP
    return True


def validate_keyword_exists(header, key):
    """Verify that the keyword exists"""
    value = header.get(key)
    if value is None:
        msg = f'Expected keyword "{key}" is not present'
        raise ValidationError(msg)
    return True


def validate_keyword_value(header, key, expected):
    from numina.exceptions import ValidationError

    validate_keyword_exists(header, key)
    value = header.get(key)

    if value != expected:
        msg = f'Keyword "{key}" has value "{value}" != "{expected}"'
        raise ValidationError(msg)


def validate_keyword_any_value(header, key, any_expected):
    """Validate that keyword has any of allowed values"""
    from numina.exceptions import ValidationError

    validate_keyword_exists(header, key)
    value = header.get(key)

    for expected in any_expected:
        if value == expected:
            break
    else:
        msg = f'Keyword "{key}" has value "{value}" not in "{any_expected}"'
        raise ValidationError(msg)


def convert_headers(hdulist):
    headers = [convert_header(hdu.header) for hdu in hdulist]
    return headers


def convert_header(header):
    hdu_v = {}
    hdu_c = {}
    hdu_o = []
    hdu_repr = {'values': hdu_v, 'comments': hdu_c, 'ordering': hdu_o}

    for card in header.cards:
        key = card.keyword
        value = card.value
        comment = card.comment
        hdu_v[key] = value
        hdu_c[key] = comment
        hdu_o.append(key)

    return hdu_repr


def check_null(obj, level=None):
    return True


def check_invalid(obj, level=None):
    raise ValidationError


class Checker(object):
    def __init__(self, validator):
        super(Checker, self).__init__()
        self.validator = validator

    def check(self, obj, level=None):

        return self.check_post(obj, level=level)

    def __call__(self, hdulist, level=None):
        return self.check(hdulist, level=level)

    def check_post(self, hdulist, level=None):
        return True


class ImageChecker(Checker):
    def __init__(self, validator):
        super(ImageChecker, self).__init__(validator)

    def check(self, hdulist, level=None):
        dheaders = convert_headers(hdulist)
        self.check_dheaders(dheaders, level=level)
        super(ImageChecker, self).check_post(hdulist, level=level)
        return True

    def check_dheaders(self, dheaders, level=None):
        # Check with json schema
        # convert header to dict
        try:
            self.validator.validate(dheaders)
        except jsonschema.exceptions.ValidationError:
            raise

        if len(dheaders) == 2:
            values_fibers = dheaders[1]['values']
            values_primary = dheaders[0]['values']
            check_header_additional(values_primary, values_fibers)


class StructChecker(Checker):
    def __init__(self, validator):
        super(StructChecker, self).__init__(validator)

    def check(self, obj, level=None):
        self.validator.validate(obj)
        self.check_post(obj, level=level)
        return True


class BaseChecker(ImageChecker):
    def __init__(self, validator):
        super(BaseChecker, self).__init__(validator)


# TODO: insert all subschemas in the general schema
_sub_schema_rss = {
    "oneOf": [
    {
        "type": "object",
        "properties": {
            # "NAXIS1": {"const": 4300},
            "NAXIS2": {"const": 623},
            "INSMODE": {"const": "LCB"}
        }
    },
    {
        "type": "object",
        "properties": {
            # "NAXIS1": {"const": 4300},
            "NAXIS2": {"const": 644},
            "INSMODE": {"const": "MOS"}
        }
    }
    ]
}

_sub_schema_bias = {
    "type": "object",
    "properties": {
        "OBJECT": {"const": "BIAS"},
        "OBSMODE": {"const": "MegaraBiasImage"},
        "IMAGETYP": {"const": "IMAGE_BIAS"},
        "EXPTIME": {"type": "number", "maximum": 0},
        "DARKTIME": {"type": "number", "maximum": 0}
    }
}


_sub_schema_master_bpm = {
    "type": "object",
    "properties": {
        "NUMTYPE": {"enum": ['MasterBias', 'MASTER_BPM']}
    }
}

_sub_schema_master_bias = {
    "type": "object",
    "properties": {
        # "OBJECT": {"const": "BIAS"},
        "OBSMODE": {"const": "MegaraBiasImage"},
        "IMAGETYP": {"const": "MASTER_BIAS"},
        "EXPTIME": {"type": "number", "maximum": 0},
        "DARKTIME": {"type": "number", "maximum": 0},
        "NUMTYPE": {"enum": ['MasterBias', 'MASTER_BIAS']}
    }
}


_sub_schema_dark = {
    "type": "object",
    "properties": {
        "OBSMODE": {"const": "MegaraDarkImage"},
        "IMAGETYP": {"const": "IMAGE_DARK"},
    }
}


_sub_schema_master_dark = {
    "type": "object",
    "properties": {
        "OBSMODE": {"const": "MegaraDarkImage"},
        "IMAGETYP": {"const": "MASTER_DARK"},
    }
}


class ExtChecker(BaseChecker):
    def __init__(self, schema, sub_schemas, n_ext=None):
        super(ExtChecker, self).__init__(schema)
        self.n_ext = n_ext
        self.sub_schemas = sub_schemas

    def check_dheaders(self, dheaders, level=None):

        try:
            super(ExtChecker, self).check_dheaders(dheaders, level=level)
        except jsonschema.exceptions.ValidationError:
            pass
        # Image must have only one extension
        if self.n_ext is not None:
            if len(dheaders) != self.n_ext:
                msg = f'image has not expected number of HDUs ({self.n_ext})'
                raise ValueError(msg)

        for sub_schema in self.sub_schemas:
            try:
                if isinstance(sub_schema, str):
                    url, fragment = self.validator.resolver.resolve(sub_schema)
                else:
                    fragment = sub_schema

                jsonschema.validate(dheaders[0]['values'], schema=fragment)
            except jsonschema.exceptions.ValidationError:
                raise


class FlatImageChecker(ExtChecker):
    def __init__(self, schema):

        _sub_schema_flat = {
            "type": "object",
            "properties": {
                "OBSMODE": {"enum": [
                    "MegaraFiberFlatImage", "MegaraTraceMap", "MegaraModelMap", "MegaraSuccess"]
                },
                "IMAGETYP": {"const": "IMAGE_FLAT"},
            }
        }

        super(FlatImageChecker, self).__init__(schema, ["#/definitions/raw_hdu_values", _sub_schema_flat], n_ext=2)

    def check_post(self, hdulist, level=None):
        """Additional checks"""
        self.check_post_level1(hdulist)

    def check_post_level1(self, hdulist):
        """Additional checks"""
        hdr = hdulist[0].header
        # Flat must have inc LAMPS-ON
        # Flat must have comp LAMPS-OFF
        lamp_i_s = (hdr['LAMPI1S'] or hdr['LAMPI1S'])
        if not lamp_i_s:
            msg = 'all incandescent lamps are OFF'
            raise ValidationError(msg)
        lamp_s_s = True
        for idx in range(1, 6):
            label = f'LAMPS{idx}S'
            lamp_s_s = lamp_s_s and hdr[label]
        if lamp_s_s:
            msg = 'some comparation lamps are ON'
            raise ValidationError(msg)


class CompImageChecker(ExtChecker):
    def __init__(self, schema):

        _sub_schema_comp = {
            "type": "object",
            "properties": {
                "OBSMODE": {"enum": ["MegaraArcCalibration", "MegaraSuccess"]},
                "IMAGETYP": {"const": "IMAGE_COMP"},
            }
        }

        super(CompImageChecker, self).__init__(schema, ["#/definitions/raw_hdu_values", _sub_schema_comp], n_ext=2)

    def check_post(self, hdulist, level=None):
        """Additional checks"""
        self.check_post_level1(hdulist)

    def check_post_level1(self, hdulist):
        """Additional checks"""
        hdr = hdulist[0].header
        # Flat must have all inc LAMPS-OFF
        # Flat must have some comp LAMPS-ON
        lamp_i_s = (hdr['LAMPI1S'] or hdr['LAMPI1S'])
        if lamp_i_s:
            msg = 'some incandescent lamps are ON'
            raise ValidationError(msg)
        lamp_s_s = False
        for idx in range(1, 6):
            label = f'LAMPS{idx}S'
            lamp_s_s = lamp_s_s or hdr[label]
        if not lamp_s_s:
            msg = 'all comparation lamps are OFF'
            raise ValidationError(msg)


class TargetImageChecker(ExtChecker):
    def __init__(self, schema):

        _sub_schema_target = {
            "type": "object",
            "properties": {
            }
        }

        super(TargetImageChecker, self).__init__(schema,
            ["#/definitions/raw_hdu_values", _sub_schema_target], n_ext=2
        )

    def check_post(self, hdulist, level=None):
        """Additional checks"""
        self.check_post_level1(hdulist)

    def check_post_level1(self, hdulist):
        """Additional checks"""
        hdr = hdulist[0].header
        # Target must have all inc LAMPS-OFF
        # Target must have all comp LAMPS-ON
        lamp_i_s = (hdr['LAMPI1S'] or hdr['LAMPI1S'])
        if lamp_i_s:
            msg = 'some incandescent lamps are ON'
            raise ValidationError(msg)
        lamp_s_s = False
        for idx in range(1, 6):
            label = f'LAMPS{idx}S'
            lamp_s_s = lamp_s_s or hdr[label]
        if lamp_s_s:
            msg = 'some comparation lamps are ON'
            raise ValidationError(msg)


class MasterFlatRSSChecker(ExtChecker):
    def __init__(self, schema):
        super(MasterFlatRSSChecker, self).__init__(schema,
            [_sub_schema_rss], n_ext=3
        )


class MasterSensitivityChecker(ExtChecker):
    def __init__(self, schema):
        super(MasterSensitivityChecker, self).__init__(schema,
            ["#/definitions/spec_hdu_values", "#/definitions/sensitivity_values"], n_ext=1
        )


def check_header_additional(values_primary, values_fibers):
    """Additional checks than can't be done with schema"""

    if values_primary['INSMODE'] != values_fibers['INSMODE']:
        raise ValueError('insmode in PRIMARY != insmode in FIBERS')

    if values_fibers['INSMODE'] == 'LCB':
        rbundles = [0, 93, 94, 95, 96, 97, 98, 99, 100]
    else:
        rbundles = range(1, 92 + 1)

    nfibers = values_fibers['NFIBERS']
    nbundles = values_fibers['NBUNDLES']

    for idbundle in rbundles:
        # types are check in the json schema
        for stype in ['P', "I", "T", "X", "Y", "O", "E"]:
            keyname = f"BUN{idbundle:03d}_{stype}"
            if keyname not in values_fibers:
                raise ValueError(f"keyname {keyname} not in values_fibers")

    for idfiber in range(1, nfibers + 1):
        # types are check in the json schema
        for stype in ['A', "D", "R", "X", "Y", "B"]:
            keyname = f"FIB{idfiber:03d}_{stype}"
            if keyname not in values_fibers:
                msg = f"keyname {keyname} not in values_fibers"
                raise ValueError(msg)

        for stype in ["N"]:
            keyname = f"FIB{idfiber:03d}_{stype}"
            if keyname not in values_fibers:
                msg = f"keyword {keyname} not in values_fibers"
                print(msg)
                # raise ValueError(msg)


class CheckAsDatatype(object):
    """Collection of schemas for validation"""
    def __init__(self):

        image_schema_path = "baseimage.json"
        json_schema_path = "basestruct.json"

        data_image = pkgutil.get_data('megaradrp.schemas', image_schema_path)
        data_json = pkgutil.get_data('megaradrp.schemas', json_schema_path)
        schema_image = json.load(StringIO(data_image.decode('utf8')))
        schema_json = json.load(StringIO(data_json.decode('utf8')))

        ValClass = jsonschema.validators.validator_for(schema_image)
        self.validator_image = ValClass(schema_image)
        ValClass = jsonschema.validators.validator_for(schema_json)
        self.validator_json = ValClass(schema_json)

        raw_checker = ExtChecker(self.validator_image, ["#/definitions/raw_hdu_values"])
        proc_checker = ExtChecker(self.validator_image, ["#/definitions/proc_hdu_values"])
        rss_checker = ExtChecker(self.validator_image, [_sub_schema_rss])
        spec_checker = ExtChecker(self.validator_image, ["#/definitions/spec_hdu_values"])
        sens_checker = MasterSensitivityChecker(self.validator_image)
        struct_checker = StructChecker(self.validator_json)

        _megara_checkers = {}
        _megara_checkers[MegaraDataType.UNKNOWN] = check_null
        _megara_checkers[MegaraDataType.IMAGE_RAW] = raw_checker
        _megara_checkers[MegaraDataType.IMAGE_BIAS] = ExtChecker(self.validator_image, ["#/definitions/raw_hdu_values", _sub_schema_bias], n_ext=1)
        _megara_checkers[MegaraDataType.IMAGE_DARK] = ExtChecker(self.validator_image, ["#/definitions/raw_hdu_values", _sub_schema_dark], n_ext=1)
        _megara_checkers[MegaraDataType.IMAGE_SLITFLAT] = raw_checker
        _megara_checkers[MegaraDataType.IMAGE_FLAT] = FlatImageChecker(self.validator_image)
        _megara_checkers[MegaraDataType.IMAGE_COMP] = CompImageChecker(self.validator_image)
        #
        _megara_checkers[MegaraDataType.IMAGE_TWILIGHT] = raw_checker
        _megara_checkers[MegaraDataType.IMAGE_TEST] = raw_checker
        _megara_checkers[MegaraDataType.IMAGE_TARGET] = TargetImageChecker(self.validator_image)
        #
        _megara_checkers[MegaraDataType.IMAGE_PROCESSED] = proc_checker
        _megara_checkers[MegaraDataType.MASTER_BPM] = ExtChecker(self.validator_image, [
            "#/definitions/proc_hdu_values", _sub_schema_master_bpm
        ], n_ext=1)
        _megara_checkers[MegaraDataType.MASTER_BIAS] = ExtChecker(self.validator_image, ["#/definitions/proc_hdu_values", _sub_schema_master_bias], n_ext=1)
        _megara_checkers[MegaraDataType.MASTER_DARK] = ExtChecker(self.validator_image, ["#/definitions/proc_hdu_values", _sub_schema_master_dark], n_ext=1)
        _megara_checkers[MegaraDataType.MASTER_SLITFLAT] = proc_checker
        _megara_checkers[MegaraDataType.DIFFUSE_LIGHT] = proc_checker
        #
        _megara_checkers[MegaraDataType.RSS_PROCESSED] = rss_checker
        _megara_checkers[MegaraDataType.RSS_WL_PROCESSED] = rss_checker
        _megara_checkers[MegaraDataType.MASTER_FLAT] = MasterFlatRSSChecker(self.validator_image)
        _megara_checkers[MegaraDataType.MASTER_TWILIGHT] = rss_checker

        _megara_checkers[MegaraDataType.SPEC_PROCESSED] = spec_checker
        _megara_checkers[MegaraDataType.MASTER_SENSITIVITY] = sens_checker

        _megara_checkers[MegaraDataType.STRUCT_PROCESSED] = struct_checker
        _megara_checkers[MegaraDataType.TRACE_MAP] = struct_checker
        _megara_checkers[MegaraDataType.MODEL_MAP] = struct_checker
        _megara_checkers[MegaraDataType.WAVE_CALIB] = struct_checker

        self._megara_checkers = _megara_checkers

    def __call__(self, datatype):
        return self._megara_checkers[datatype]


check_as_datatype = CheckAsDatatype()



#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy


def atleast_2d_last(*arys):
    """Equivalent to atleast_2d, adding the newaxis at the end"""
    res = []
    for ary in arys:
        ary = numpy.asanyarray(ary)
        if len(ary.shape) == 0:
            result = ary.reshape(1, 1)
        elif len(ary.shape) == 1:
            result = ary[:, numpy.newaxis]
        else:
            result = ary
        res.append(result)
    if len(res) == 1:
        return res[0]
    else:
        return res



#
# Copyright 2017-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Interpolation method based on:
'Hex-Splines: A Novel Spline Family for Hexagonal Lattices'
van de Ville et al. IEEE Transactions on Image Processing 2004, 13, 6
"""

from __future__ import print_function

import numpy as np
from scipy import signal
import astropy.units as u
import astropy.wcs
from numina.frame.utils import copy_img

from megaradrp.instrument.focalplane import FocalPlaneConf
# from megaradrp.datamodel import MegaraDataModel
from megaradrp.core.utils import atleast_2d_last
import megaradrp.processing.fixrss as fixrss
import megaradrp.processing.hexgrid as hg
import megaradrp.processing.hexspline as hspline
import megaradrp.instrument.constants as cons

# Helper function for equivalence conversion
GTC_PLATESCALE = u.plate_scale(cons.GTC_FC_A_PLATESCALE)
# Size scale of the spaxel grid in arcseconds
HEX_SCALE = cons.SPAXEL_SCALE.to(u.arcsec, GTC_PLATESCALE).value


def calc_matrix_from_fiberconf(fpconf, refid=614):
    """
    Compute hexagonal grid matrix from FocalPlaneConf

    Parameters
    ----------
    fpconf : megaradrp.instrument.focalplane.FocalPlaneConf
    refid : int
        fiber ID of reference fiber for grid coordinates
    Returns
    -------

    """

    # TODO: This should be in FIBERCONFS...
    spos1_x = []
    spos1_y = []
    for fiber in fpconf.connected_fibers():
        spos1_x.append(fiber.x)
        spos1_y.append(fiber.y)
    spos1_x = np.asarray(spos1_x)
    spos1_y = np.asarray(spos1_y)

    # FIBER in LOW LEFT corner is 614
    ref_fiber = fpconf.fibers[refid]
    minx, miny = ref_fiber.x, ref_fiber.y
    if fpconf.funit == 'arcsec':
        # arcsec
        ascale = HEX_SCALE
    else:
        # mm
        # fpconf.funit == 'mm'
        ascale = cons.SPAXEL_SCALE.to(u.mm).value
    ref = minx / ascale, miny / ascale
    rpos1_x = (spos1_x - minx) / ascale
    rpos1_y = (spos1_y - miny) / ascale
    r0l_1 = np.array([rpos1_x, rpos1_y])
    return r0l_1, ref


def create_cube(r0l, zval, p=1, target_scale=1.0):
    """

    Parameters
    ----------
    r0l
    zval
    p : {1, 2}
    target_scale : float, optional

    Returns
    -------

    Raises
    ------
    ValueError
        If `p` > 2

    """
    # geometry
    # Interpolation method. Allowed values are:
    # P = 1 NN
    # P = 2 Linear
    if p > 2:
        raise ValueError('p > 2 not implemented')

    rr1 = target_scale * np.array([[1.0, 0], [0, 1]])  # Unit scale

    # compute extremes of hexgrid to rectangular grid
    # with pixel size 'scale'

    (i1min, i1max), (j1min, j1max) = hg.hexgrid_extremes(r0l, target_scale)

    # Rectangular grid
    mk1 = np.arange(i1min, i1max + 1)
    mk2 = np.arange(j1min, j1max + 1)
    crow = len(mk1)
    ccol = len(mk2)
    # Result image
    # Add third last axis
    zval2 = atleast_2d_last(zval)
    # disp axis is last axis...
    dk = np.zeros((crow, ccol, zval2.shape[-1]))
    # print('result shape is ', dk.shape)
    # r1k = rr1 @ sk
    sk = np.flipud(np.transpose([np.tile(mk1, len(mk2)), np.repeat(mk2, len(mk1))]).T)  # x y
    r1k = np.dot(rr1, sk)

    # Prefiltering
    # For p = 1, prefilter coefficients with p = 1, coeff = 1
    # For p = 2, prefilter coefficients with p = 2, coeff = 1
    # No prefiltering in zval2 is required if p <= 2

    rbs = hspline.rescaling_kernel(p, scale=target_scale)

    # Loop to compute integrals...
    for s, r in zip(sk.T, r1k.T):
        allpos = -(r0l - r[:, np.newaxis])
        we = np.abs((rbs.ev(allpos[1], allpos[0])))
        dk[s[1] - i1min, s[0] - j1min] = np.sum(we[:, np.newaxis] * zval2, axis=0)

    # Postfiltering
    # For p = 1, final image in NN, postfilter coefficients with n = 1
    # For p = 2, final image is linear, postfilter coefficients with n = 3
    #
    if p == 1:
        # Coefficients post filtering to n = 2 * p - 1 == 1
        cpk = dk
        # Nearest-neighbor samples equal to coefficients
        img = cpk
    elif p == 2:
        # Coefficients post filtering to n = 2 * p - 1 == 3
        cpk = np.zeros_like(dk)
        # last axis
        for k in range(dk.shape[-1]):
            cpk[..., k] = signal.cspline2d(dk[..., k])
        # Linear samples equal to coefficients
        img = cpk
    else:
        raise ValueError('p > 2 not implemented')

    return img


def create_cube_from_array(rss_data, fiberconf, p=1, target_scale_arcsec=1.0, conserve_flux=True):
    """
    Create a cube array from a 2D or 1D array and focal plane configuration

    Parameters
    ----------
    rss_data
    fiberconf : megaradrp.instrument.focalplane.FocalPlaneConf
    p : {1, 2}
    target_scale_arcsec : float
    conserve_flux : bool

    Returns
    -------
    np.ndarray

    """

    target_scale = target_scale_arcsec / HEX_SCALE
    conected = fiberconf.connected_fibers()
    rows = [conf.fibid - 1 for conf in conected]

    rss_data = atleast_2d_last(rss_data)

    region = rss_data[rows, :]

    r0l, _ = calc_matrix_from_fiberconf(fiberconf)
    cube_data = create_cube(r0l, region[:, :], p, target_scale)

    if conserve_flux:
        # scale with areas
        cube_data *= (target_scale ** 2 / hg.HA_HEX)
    # Move axis to put WL first
    # so that is last in FITS
    result = np.moveaxis(cube_data, 2, 0)
    result.astype('float32')
    return result


def create_cube_from_rss(rss, p=1, target_scale_arcsec=1.0, conserve_flux=True):
    """
    Create a cube HDUlist from a RSS HDUList

    Parameters
    ----------
    rss : fits.HDUList
    p : {1, 2}
    target_scale_arcsec : float, optional
    conserve_flux : bool, optional

    Returns
    -------
    fits.HDUList
    """

    fiberconf = FocalPlaneConf.from_img(rss)
    result_arr = create_cube_from_array(
        rss[0].data, fiberconf, p=p,
        target_scale_arcsec=target_scale_arcsec,
        conserve_flux=conserve_flux
    )

    cube = copy_img(rss)
    cube[0].data = result_arr

    sky_header = rss['FIBERS'].header.copy()
    spec_header = rss[0].header
    # Update values of sky WCS
    # CRPIX1, CRPIX2
    # CDELT1, CDELT2
    # minx, miny
    # After shifting the array
    # refpixel is -i1min, -j1min
    target_scale = target_scale_arcsec / HEX_SCALE
    r0l, (refx, refy) = calc_matrix_from_fiberconf(fiberconf)
    (i1min, i1max), (j1min, j1max) = hg.hexgrid_extremes(r0l, target_scale)
    crpix_x = -refx / target_scale - j1min
    crpix_y = -refy / target_scale - i1min
    # Map the center of original field
    sky_header['CRPIX1'] = crpix_x
    sky_header['CRPIX2'] = crpix_y
    sky_header['CDELT1'] = -target_scale_arcsec / 3600.0
    sky_header['CDELT2'] = target_scale_arcsec / 3600.0

    # Merge headers
    # 2D from FIBERS
    # WL from PRIMARY
    merge_wcs(sky_header, spec_header, out=cube[0].header)

    # done
    return cube


def merge_wcs(hdr_sky, hdr_spec, out=None):
    """Merge sky WCS with spectral WCS

    Works only with main WCS and B WCS

    """
    if out is None:
        hdr = hdr_spec.copy()
    else:
        hdr = out

    allw = astropy.wcs.find_all_wcs(hdr_spec)
    wcsnames = [w.wcs.alt for w in allw]
    for ss in wcsnames:
        merge_wcs_alt(hdr_sky, hdr_spec, hdr, spec_suffix=ss)
    return hdr


def merge_wcs_alt(hdr_sky, hdr_spec, out, spec_suffix=' '):
    """Merge sky WCS with spectral WCS"""

    hdr = out
    if spec_suffix == ' ':
        sf = ''
    else:
        sf = spec_suffix
    # Extend header for third axis
    c_crpix = 'Pixel coordinate of reference point'
    c_cunit = 'Units of coordinate increment and value'
    wcsname_s = f'WCSNAME{sf}'
    if wcsname_s in hdr:
        prev = wcsname_s
    else:
        prev = f'CTYPE1{sf}'

    hdr.set(f'WCSAXES{sf}', value=3, before=prev)
    if sf != '':
        hdr.set(f'WCSNAME{sf}', value='', after='PC3_3')
        hdr.set(f'CTYPE1{sf}', value='', after=f'WCSNAME{sf}')
        hdr.set(f'CRPIX1{sf}', value=1.0, after=f'CTYPE1{sf}')
        hdr.set(f'CRVAL1{sf}', value=0.0, after=f'CRPIX1{sf}')
        hdr.set(f'CDELT1{sf}', value=1.0, after=f'CRVAL1{sf}')
    hdr.set(f'CUNIT1{sf}', value='deg', comment=c_cunit, after=f'CDELT1{sf}')
    hdr.set(f'CTYPE2{sf}', after=f'CUNIT1{sf}')
    if sf != '':
        hdr.set(f'CRPIX2{sf}', value=1.0, after=f'CTYPE2{sf}')
        hdr.set(f'CRVAL2{sf}', value=0.0, after=f'CRPIX2{sf}')
        hdr.set(f'CDELT2{sf}', value=1.0, after=f'CRVAL2{sf}')
    hdr.set(f'CUNIT2{sf}', value='deg', comment=c_cunit, after=f'CDELT2{sf}')
    hdr.set(f'CRPIX2{sf}', value=1, comment=c_crpix, after=f'CTYPE2{sf}')
    hdr.set(f'CTYPE3{sf}', after=f'CUNIT2{sf}')
    hdr.set(f'CRPIX3{sf}', value=1, comment=c_crpix, after=f'CTYPE3{sf}')
    hdr.set(f'CRVAL3{sf}', after=f'CRPIX3{sf}')
    hdr.set(f'CDELT3{sf}', after=f'CRVAL3{sf}')
    hdr.set(f'CUNIT3{sf}', comment=c_cunit, after=f'CDELT3{sf}')
    c_pc = 'Coordinate transformation matrix element'
    hdr.set(f'PC1_1{sf}', value=1.0, comment=c_pc, after=f'CUNIT3{sf}')
    hdr.set(f'PC1_2{sf}', value=0.0, comment=c_pc, after=f'PC1_1{sf}')
    hdr.set(f'PC2_1{sf}', value=0.0, comment=c_pc, after=f'PC1_2{sf}')
    hdr.set(f'PC2_2{sf}', value=1.0, comment=c_pc, after=f'PC2_1{sf}')
    hdr.set(f'PC3_3{sf}', value=1.0, comment=c_pc, after=f'PC2_2{sf}')

    # Mapping, which keyword comes from each header
    mappings = [('CRPIX3', 'CRPIX1', sf, 0),
                ('CDELT3', 'CDELT1', sf, 0),
                ('CRVAL3', 'CRVAL1', sf, 0),
                ('CTYPE3', 'CTYPE1', sf, 0),
                ('CRPIX1', 'CRPIX1', '', 1),
                ('CDELT1', 'CDELT1', '', 1),
                ('CRVAL1', 'CRVAL1', '', 1),
                ('CTYPE1', 'CTYPE1', '', 1),
                ('CUNIT3', 'CUNIT1', sf, 0),
                ('PC1_1', 'PC1_1', '', 1),
                ('PC1_2', 'PC1_2', '', 1),
                ('CRPIX2', 'CRPIX2', '', 1),
                ('CDELT2', 'CDELT2', '', 1),
                ('CRVAL2', 'CRVAL2', '', 1),
                ('CTYPE2', 'CTYPE2', '', 1),
                ('CUNIT2', 'CUNIT2', '', 1),
                ('PC2_1', 'PC2_1', '', 1),
                ('PC2_2', 'PC2_2', '', 1),
                ('LONPOLE', 'LONPOLE', '', 1),
                ('LATPOLE', 'LATPOLE', '', 1),
                ('RADESYS', 'RADESYS', '', 1),
                ('EQUINOX', 'EQUINOX', '', 1),
                ('WCSNAME', 'WCSNAME', sf, 0),
                ('specsys', 'SPECSYS', sf, 0),
                ('ssysobs', 'SSYSOBS', sf, 0),
                ('velosys', 'VELOSYS', sf, 0)
                ]

    hdr_in = dict()
    hdr_in[0] = hdr_spec
    hdr_in[1] = hdr_sky

    for dest, orig, key, idx in mappings:
        hdr_orig = hdr_in[idx]
        korig = orig + key
        kdest = dest + sf
        try:
            hdr[kdest] = hdr_orig[korig], hdr_orig.comments[korig]
        except KeyError:
            # Ignoring errors. Copy only if keyword exists
            pass

    return hdr


def main(args=None):
    import argparse
    import astropy.io.fits as fits

    # parse command-line options
    parser = argparse.ArgumentParser(prog='convert_rss_cube')
    # positional parameters

    methods = {'nn': 1, 'linear': 2}

    parser.add_argument("rss",
                        help="RSS file with fiber traces",
                        type=argparse.FileType('rb'))
    parser.add_argument('-p', '--pixel-size', type=float, default=0.3,
                        metavar='PIXEL_SIZE',
                        help="Pixel size in arc seconds")
    parser.add_argument('-o', '--outfile', default='cube.fits',
                        help="Name of the output cube file")
    parser.add_argument('-d', '--disable-scaling', action='store_true',
                        help="Disable flux conservation")
    parser.add_argument('-m', '--method', action='store', choices=['nn', 'linear'],
                        default='nn', help="Method of interpolation")
    parser.add_argument('--wcs-pa-from-header', action='store_true',
                        help="Use PA angle from header", dest='pa_from_header')
    parser.add_argument('--fix-missing', action='store_true',
                        help="Interpolate missing fibers")

    args = parser.parse_args(args=args)

    target_scale = args.pixel_size  # Arcsec
    p = methods[args.method]
    print(f'interpolation method is "{args.method}"')
    print('target scale is', target_scale, 'arcsec')
    conserve_flux = not args.disable_scaling

    with fits.open(args.rss) as rss:
        if not args.pa_from_header:
            # Doing it here so the change is propagated to
            # all alternative coordinates
            print('recompute WCS from IPA')
            ipa = rss['PRIMARY'].header['IPA']
            rss['FIBERS'].header = fixrss.recompute_wcs(rss['FIBERS'].header, ipa=ipa)
        if args.fix_missing:
            fibid = 623
            print(f'interpolate fiber {fibid}')
            rss = fixrss.fix_missing_fiber(rss, fibid)

        cube = create_cube_from_rss(rss, p, target_scale, conserve_flux=conserve_flux)

    cube.writeto(args.outfile, overwrite=True)


if __name__ == '__main__':

    main()



import pytest
import numpy as np
import math

from ..hexspline import rescaling_kernel, hexspline2

@pytest.mark.parametrize("p", [1, 2])
@pytest.mark.parametrize("scale", [0.5, 0.8, 1])
def test_rescaling_kernel_normalization(p, scale):
    rbs = rescaling_kernel(p, scale=scale)
    expected_area = math.sqrt(3) / 2
    area = rbs.integral(-3.0, 3.0, -3.0, 3.0)
    assert np.allclose(area, expected_area, rtol=1e-2)


def test_hexspline2():
    x = [0.45, -0.2, 1, 0, 0.68, 1.2]
    y = [-0.8, -0.3, 0.5, 0, 0.2, 0.5]
    expected_res = [5.60769515e-02, 5.78341925e-01, 6.40987562e-16, 1.00000000e+00,
       2.11842907e-01, 7.69185075e-16]
    res = hexspline2(x, y)
    assert np.allclose(res, expected_res, rtol=1e-2)


import pytest

from ..wheel import VPHWheel


class Vph(object):
    def __init__(self, name):
        self.name = name


@pytest.fixture
def wheel_dev():
    wheel = VPHWheel(3)
    for idx in range(3):
        wheel.put_in_pos(Vph(idx), idx)
    return wheel


def test_wheel(wheel_dev):
    curr = wheel_dev.current()
    assert isinstance(curr, Vph)
    assert curr.name == 0
    assert wheel_dev.pos() == 0


#
# Copyright 2011-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

from numina.core import BaseRecipe
from numina.core import DataFrame
from numina.types.qc import QC
from numina.core.requirements import ObservationResultRequirement
import numina.util.flow as flowmod


import megaradrp.core.correctors as cor
from megaradrp.datamodel import MegaraDataModel


class MegaraBaseRecipe(BaseRecipe):
    """Base class for all MEGARA Recipes

    Attributes
    ----------

    obresult : ObservationResult, requirement
    logger :
         recipe logger

    datamodel : MegaraDataModel

    """

    obresult = ObservationResultRequirement()
    logger = logging.getLogger('numina.recipes.megara')
    datamodel = MegaraDataModel()

    def validate_input(self, recipe_input):
        """"Validate the input of the recipe"""

        import numina.types.multitype
        # print('ATTRS', recipe_input.attrs())
        # print('STORED', recipe_input.stored())
        # print('tag_names', recipe_input.tag_names())

        # Find reference tags
        ref_tags = {}
        for key, val in recipe_input.attrs().items():
            if key == 'obresult':
                ref_tags = val.tags
                break

        # super(MegaraBaseRecipe, self).validate_input(recipe_input)
        # check all the rest against reference tags
        stored = recipe_input.stored()
        attrs = recipe_input.attrs()
        val_results = []
        for key, val in attrs.items():
            if val is None:
                continue
            # If we evaluate query_expr with ref_tags and the tags from the object
            # The result must be true
            req = stored[key]
            rtype = req.type
            # TODO: develop a method to select images that are valid
            # for different filters or insmodes; perhaps a ANY keyword
            tags = rtype.extract_tags(val)
            # FIXME: this should be handled by the type, not with a special case
            if isinstance(rtype, numina.types.multitype.MultiType):
                # get query from first node
                query_expr = rtype.node_type[0].query_expr
            else:
                query_expr = rtype.query_expr

            q2 = query_expr.fill_placeholders(**ref_tags)
            self.logger.debug('type %s with tags %s, expr %s', rtype, tags, q2)
            is_valid = q2.eval(**tags)
            if not is_valid:
                val_results.append((key, q2, tags, ref_tags))
        msg = 'invalid {} with expression {} and tags {} and obs_tags {}'
        for key, q2, tags, ref_tags in val_results:
            self.logger.error(msg.format(key, q2, tags, ref_tags))
        if val_results:
            raise ValueError('Validation error', val_results)


    def run_qc(self, recipe_input, recipe_result):
        """Run Quality Control checks."""
        recipe_result.qc = QC.GOOD
        return recipe_result

    def types_getter(self):
        from megaradrp.ntypes import MasterBias, MasterDark, MasterBPM, MasterSlitFlat
        from megaradrp.ntypes import DiffuseLightCorrection
        imgtypes = [MasterBPM, MasterBias, MasterDark, MasterSlitFlat, DiffuseLightCorrection]
        getters = [cor.get_corrector_bpm, cor.get_corrector_bias,
                   [cor.get_corrector_dark, cor.get_corrector_gain],
                   cor.get_corrector_slit_flat,
                   cor.get_corrector_diffuse_light,
                   ]
        return imgtypes, getters

    def get_filters(self, imgtypes, getters):
        import collections
        used_getters = []
        for rtype, getter in zip(imgtypes, getters):
            self.logger.debug('get_filters, %s  %s', rtype, getter)
            if rtype is None:
                # Unconditional
                if isinstance(getter, collections.Iterable):
                    used_getters.extend(getter)
                else:
                    used_getters.append(getter)
            else:
                # Search
                for key, val in self.RecipeInput.stored().items():
                    if isinstance(val.type, rtype):
                        if isinstance(getter, collections.Iterable):
                            used_getters.extend(getter)
                        else:
                            used_getters.append(getter)
                        break
                else:
                    pass
        return used_getters

    def init_filters_generic(self, rinput, getters_seq, ins):

        meta = self.gather_info(rinput)
        self.logger.debug('obresult info')
        for entry in meta['obresult']:
            self.logger.debug('frame info is %s', entry)

        reduction_flows = []
        for getters in getters_seq:
            correctors = [getter(rinput, meta, ins, self.datamodel) for getter in getters]
            reduction_flow = flowmod.SerialFlow(correctors)
            reduction_flows.append(reduction_flow)

        return reduction_flows

    def init_filters(self, rinput, ins):
        imgtypes1 = [None]
        getters1 = [[cor.get_corrector_overscan, cor.get_corrector_trimming]]
        getters_f1 = self.get_filters(imgtypes1, getters1)
        imgtypes2, getters2 = self.types_getter()
        getters_f2 = self.get_filters(imgtypes2, getters2)

        getters_seq = [getters_f1, getters_f2]
        return self.init_filters_generic(rinput, getters_seq, ins)

    def gather_info(self, recipeinput):
        klass = recipeinput.__class__
        metadata = {}
        for key in klass.stored():
            val = getattr(recipeinput, key)
            if isinstance(val, DataFrame):
                metadata[key] = self.datamodel.gather_info_dframe(val)
            # elif isinstance(val, ObservationResult):
            elif hasattr(val, 'frames'):
                metadata[key] = self.datamodel.gather_info_oresult(val)
            else:
                pass
        return metadata



import pytest
import astropy.io.fits as fits

from megaradrp.tests.simpleobj import create_spec_header
from ..wavecalibration import header_add_barycentric_correction


def test_add_barycentric_missing1():
    hdr = create_spec_header()
    del hdr['RADEG']

    with pytest.raises(KeyError):
        header_add_barycentric_correction(hdr, key='b')


def test_add_barycentric_missing2():
    hdr = create_spec_header()
    del hdr['DATE-OBS']

    with pytest.raises(KeyError):
        header_add_barycentric_correction(hdr, key='b')


def test_add_barycentric_missing3():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2017-08-23T21:38:30.55'
    # GTC
    hdr['OBSGEO-X'] = 5327285.0921
    hdr['OBSGEO-Y'] = -1718777.1125
    hdr['OBSGEO-Z'] = 3051786.7327

    hdr['RADEG'] = 285.481037748898
    hdr['DECDEG'] = 42.4882140636786

    with pytest.raises(TypeError):
        header_add_barycentric_correction(hdr, key='b')




import pytest

import megaradrp.datamodel as dm


@pytest.fixture(scope='module')
def focalplane(request):
    name = request.param
    fp = dm.get_fiberconf_default(name)
    return fp


@pytest.mark.parametrize("focalplane", ['MOS'], indirect=["focalplane"])
@pytest.mark.parametrize("fibid, around", [
    [372, [373, 374, 375]],
    [8, [9, 10, 11]],
    [623, [620, 621, 622]],
    [26, [23, 25, 28]],  # sky fiber
    [25, [22, 23, 24, 26, 27, 28]],  # sky fiber
    [487, [484, 485, 486, 488, 489, 490]],  # sky fiber
    [490, [487, 488, 489]]
])
def test_nearby_mos(focalplane, fibid, around):
    # Find the bundle the fiber belongs to
    result = focalplane.nearby_fibers(fibid)
    assert sorted(result) == sorted(around)


@pytest.mark.parametrize("focalplane", ['LCB'], indirect=["focalplane"])
@pytest.mark.parametrize("fibid, around", [
    [372, [250, 376, 252, 388, 374, 238]],
    [8, [12, 10]],
    [623, [184, 185, 525, 529, 620, 622]],
    [26, [23, 25, 28]],  # sky fiber
    [25, [22, 23, 24, 26, 27, 28]],  # sky fiber
    [487, [484, 485, 486, 488, 489, 490]],  # sky fiber
    [490, [487, 488, 489]]
])
def test_nearby_lcb(focalplane, fibid, around):
    # Find the bundle the fiber belongs to
    result = focalplane.nearby_fibers(fibid)
    assert sorted(result) == sorted(around)


@pytest.mark.parametrize("focalplane", ['LCB'], indirect=["focalplane"])
@pytest.mark.parametrize("fibid", [700, -1, 624])
def test_nearby_exception_lcb(focalplane, fibid):
    with pytest.raises(ValueError):
        focalplane.nearby_fibers(fibid)


@pytest.mark.parametrize("focalplane", ['MOS'], indirect=["focalplane"])
@pytest.mark.parametrize("fibid", [700, -1, 645])
def test_nearby_exception_mos(focalplane, fibid):
    with pytest.raises(ValueError):
        focalplane.nearby_fibers(fibid)



#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Typical requirements of recipes"""

from numina.core import Requirement
from numina.types.multitype import MultiType

import megaradrp.ntypes
import megaradrp.products
import megaradrp.products.modelmap


class MasterBiasRequirement(Requirement):
    """Requirement for Master Bias images"""
    def __init__(self, optional=False):
        super(MasterBiasRequirement, self).__init__(
            megaradrp.ntypes.MasterBias,
            'Master BIAS image',
            optional=optional,
            validation=True
        )


class MasterBPMRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterBPMRequirement, self).__init__(
            megaradrp.ntypes.MasterBPM,
            'Master Bad Pixel Mask',
            optional=optional,
            validation=False
        )


class MasterDarkRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterDarkRequirement, self).__init__(
            megaradrp.ntypes.MasterDark,
            'Master DARK image',
            optional=optional,
            validation=True
        )


class MasterFiberFlatRequirement(Requirement):
    def __init__(self):
        super(MasterFiberFlatRequirement, self).__init__(
            megaradrp.ntypes.MasterFiberFlat,
            'Master fiber flat calibration',
            validation=True
        )


class MasterSlitFlatRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterSlitFlatRequirement,
              self).__init__(megaradrp.ntypes.MasterSlitFlat,
                             'Master slit flat calibration',
                             optional=optional
                             )


class MasterTwilightRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterTwilightRequirement,
              self).__init__(megaradrp.ntypes.MasterTwilightFlat,
                             'Master twlight flat calibration',
                             optional=optional
                             )


class MasterTraceMapRequirement(Requirement):
    def __init__(self):
        super(MasterTraceMapRequirement, self).__init__(
            megaradrp.products.TraceMap,
            'Trace information of the Apertures',
            validation=True
        )


class MasterAperturesRequirement(Requirement):
    def __init__(self, alias=None):
        super(MasterAperturesRequirement, self).__init__(MultiType(
            megaradrp.products.modelmap.ModelMap,
            megaradrp.products.TraceMap),
            'Apertures information for extraction',
            validation=True,
            alias=alias
        )


class WavelengthCalibrationRequirement(Requirement):
    def __init__(self):
        super(WavelengthCalibrationRequirement, self).__init__(
            megaradrp.products.WavelengthCalibration,
            'Wavelength calibration table',
            validation=True
        )


class LinesCatalogRequirement(Requirement):
    def __init__(self):
        super(LinesCatalogRequirement, self).__init__(megaradrp.ntypes.MegaraLinesCatalog, 'Catalog of lines')


class SkyRSSRequirement(Requirement):
    def __init__(self, optional=True):
        super(SkyRSSRequirement, self).__init__(
            megaradrp.ntypes.SkyRSS,
            'Row Stacked Spectra of the sky',
            optional=optional
        )


class SensitivityRequirement(Requirement):
    def __init__(self, optional=True):
        super(SensitivityRequirement,
              self).__init__(megaradrp.ntypes.MasterSensitivity,
                             'Master sensitivity for flux calibration',
                             optional=optional
                             )


class ReferenceExtinction(Requirement):
    def __init__(self, optional=True):
        super(ReferenceExtinction,
              self).__init__(megaradrp.ntypes.ReferenceExtinctionTable,
                             "Reference extinction",
                             optional=optional
                             )


class DiffuseLightRequirement(Requirement):
    def __init__(self, optional=True):
        super(DiffuseLightRequirement,
              self).__init__(megaradrp.ntypes.DiffuseLightCorrection,
                             'Diffuse light correction image',
                             optional=optional
                             )



import numpy

from megaradrp.datamodel import create_default_fiber_header
from ..sky import subtract_sky_rss


def create_rss(value, wlmap):
    import astropy.io.fits as fits
    data1 = value + numpy.zeros((623, 4300), dtype='float32')
    hdu = fits.PrimaryHDU(data1)
    hdrf = create_default_fiber_header('LCB')
    fibers = fits.ImageHDU(header=hdrf, name='FIBERS')
    rss_map = fits.ImageHDU(wlmap, name='WLMAP')
    return fits.HDUList([hdu, fibers, rss_map])


def test_subtract_sky_rss():

    wlmap = numpy.zeros((623, 4300), dtype='float32')
    wlmap[:,350:4105] = 1.0
    wlmap[622,:] = 0
    img1 = create_rss(1000, wlmap)
    img2 = create_rss(400, wlmap)

    final_img, img, sky_img = subtract_sky_rss(img1, img2)
    assert img is img1
    # In final image, regions outside WLMAP must be at zero
    assert final_img[0].data[:, 100:200].min() == 0
    assert final_img[0].data[:, 100:200].max() == 0

    assert final_img[0].data[622, :].max() == 0
    assert final_img[0].data[622, :].min() == 0



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""

import logging

from numina.core import DataFrameType
from numina.types.product import DataProductMixin
from numina.types.datatype import DataType
from numina.types.array import ArrayType, ArrayNType
from numina.types.linescatalog import LinesCatalog
from numina.exceptions import ValidationError

import megaradrp.validators as valid
from megaradrp.datatype import MegaraDataType
from megaradrp.datamodel import MegaraDataModel, QueryAttribute


_logger = logging.getLogger(__name__)


def validate_fiber_ext(header_f):
    _logger.debug('validate fiber extension')


class Point2D(ArrayNType):
    """A type of 2D cartesian coordinates"""
    def __init__(self, *args, **kwds):
        super(ArrayNType, self).__init__(2)


class MegaraFrame(DataFrameType):
    """A processed frame"""
    DATATYPE = MegaraDataType.IMAGE_RAW
    tags_headers = {}

    def __init__(self, *args, **kwds):
        super(MegaraFrame, self).__init__(datamodel=MegaraDataModel)

    def validate_hdulist(self, hdulist):
        _logger.debug('validate MasterBias')
        checker = valid.check_as_datatype(self.DATATYPE)
        return checker(hdulist)

    def extract_tags(self, obj):
        """Extract tags from serialized file"""

        objl = self.convert(obj)
        ext = self.datamodel.extractor_map['fits']
        tags = {}

        if objl:
            with objl.open() as hdulist:
                for field in self.names_t:
                    tags[field] = ext.extract(field, hdulist)
                return tags
        else:
            return {}


class ProcessedFrame(MegaraFrame):
    """A processed frame"""
    DATATYPE = MegaraDataType.IMAGE_PROCESSED
    tags_headers = {}


class ProcessedImage(ProcessedFrame):
    """A processed image"""
    DATATYPE = MegaraDataType.IMAGE_PROCESSED


class ProcessedRSS(ProcessedFrame):
    """A processed RSS image"""
    DATATYPE = MegaraDataType.RSS_PROCESSED


class ProcessedMultiRSS(ProcessedFrame):
    """A processed RSS image not to be stored"""
    pass


class ProcessedSpectrum(ProcessedFrame):
    """A 1d spectrum"""
    DATATYPE = MegaraDataType.SPEC_PROCESSED
    pass


class ProcessedImageProduct(DataProductMixin, ProcessedImage):
    pass


class ProcessedRSSProduct(DataProductMixin, ProcessedRSS):
    pass


class ProcessedSpectrumProduct(DataProductMixin, ProcessedSpectrum):
    pass


class MegaraLinesCatalog(LinesCatalog):
    __tags__ = {'speclamp': QueryAttribute('speclamp', str), 'vph': QueryAttribute('vph', str)}
    # We are not passing the table of query_attrs in datamodel

    def name(self):
        return "LinesCatalog"


class MasterBias(ProcessedImageProduct):
    """A Master Bias image"""
    DATATYPE = MegaraDataType.MASTER_BIAS


class MasterTwilightFlat(ProcessedRSSProduct):
    __tags__ = ['insmode', 'vph', 'confid']
    DATATYPE = MegaraDataType.MASTER_TWILIGHT


class MasterDark(ProcessedImageProduct):
    """A Master Dark image"""
    DATATYPE = MegaraDataType.MASTER_DARK


class MasterFiberFlat(ProcessedRSSProduct):
    __tags__ = ['insmode', 'vph', 'confid']
    DATATYPE = MegaraDataType.MASTER_FLAT


class MasterSlitFlat(ProcessedImageProduct):
    __tags__ = ['insmode', 'vph']
    DATATYPE = MegaraDataType.MASTER_SLITFLAT


class MasterBPM(ProcessedImageProduct):
    """Bad Pixel Mask product"""
    DATATYPE = MegaraDataType.MASTER_BPM


class DiffuseLightCorrection(ProcessedImageProduct):
    """Image to correct from diffuse light"""
    pass


class MasterSensitivity(ProcessedSpectrumProduct):
    """Sensitivity correction."""
    pass


class SkyRSS(ProcessedRSS):
    """A processed RSS image"""
    pass


class ReferenceExtinctionTable(DataProductMixin, ArrayType):
    """Atmospheric Extinction."""

    def validate(self, obj):
        if obj is None:
            # None is valid
            pass
        else:
            super(ReferenceExtinctionTable, self).validate(obj)

    def convert(self, obj):
        # Support None value
        if obj is None:
            return None
        else:
            return super(ReferenceExtinctionTable, self).convert(obj)


class ReferenceSpectrumTable(DataProductMixin, ArrayType):
    """The spectrum of a reference star"""
    pass


class JSONstorage(DataType):
    def __init__(self, default=None):
        super(JSONstorage, self).__init__(ptype=dict, default=default)

    def _datatype_dump(self, obj, where):
        import json
        filename = where + '.json'

        with open(filename, 'w') as fd:
            fd.write(json.dumps(obj, sort_keys=True, indent=2,
                                separators=(',', ': ')))

        return filename

    def _datatype_load(self, obj):
        import json
        try:
            with open(obj, 'r') as fd:
                data = json.load(fd)
        except IOError as e:
            raise e
        return data


class FocusWavelength(JSONstorage):
    """Rich table with focus and wavelength"""
    pass



#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Interpolation method based on:
'Hex-Splines: A Novel Spline Family for Hexagonal Lattices'
van de Ville et al. IEEE Transactions on Image Processing 2004, 13, 6
"""

from __future__ import division

import math
import numpy as np
from megaradrp.simulation.convolution import hex_c

# Hexagon constants
M_SQRT3 = math.sqrt(3)


def hexspline1(xx, yy):
    """Hexspline of order p=1"""
    return hex_c(xx, yy, rad=1.0 / M_SQRT3, ang=0.0)


def hexspline2(x1, x2):
    """Hexspline of order p=2"""

    # This function has been created with sympy

    x1 = np.asanyarray(x1)
    x2 = np.asanyarray(x2)
    Heaviside = np.heaviside

    M_SQRT3_D_3 = M_SQRT3 / 3
    M_SQRT3_D_2 = M_SQRT3 / 2
    M_SQRT3_D_6 = M_SQRT3 / 6
    M_SQRT3_2D3 = 2 * M_SQRT3_D_3
    ref = 1 / 2

    x_p0_m1 = x2 - 1
    x_p0_p1 = x2 + 1
    x_p0_m12 = x2 - 1 / 2
    x_p0_p12 = x2 + 1 / 2

    x_p0_scaled = M_SQRT3_D_3 * x2
    x_p0_m1_scaled = M_SQRT3_D_3 * x_p0_m1
    x_p0_p1_scaled = M_SQRT3_D_3 * x_p0_p1
    x_p0_m12_scaled = M_SQRT3_D_3 * x_p0_m12
    x_p0_p12_scaled = M_SQRT3_D_3 * x_p0_p12

    aux_p1 = x1 + x_p0_m1_scaled
    aux_p1_m3 = aux_p1 - M_SQRT3_D_3

    aux_m1 = x1 - x_p0_m1_scaled
    aux_m1_p3 = x1 - x_p0_m1_scaled + M_SQRT3_D_3

    aux_mm1 = x1 - x_p0_p1_scaled
    aux_mm1_m3 = aux_mm1 - M_SQRT3_D_3
    aux_pm1 = x1 + M_SQRT3 * (x_p0_p1) / 3
    aux_pm1_p3 = aux_pm1 + M_SQRT3_D_3

    aux_p12 = x1 + x_p0_m12_scaled
    aux_p12_m2 = aux_p12 - M_SQRT3_D_2
    aux_p12_p2 = aux_p12 + M_SQRT3_D_2
    aux_p12_p6 = aux_p12 + M_SQRT3_D_6
    aux_p12_m6 = aux_p12 - M_SQRT3_D_6

    aux_m12 = x1 - x_p0_m12_scaled
    aux_m12_m6 = aux_m12- M_SQRT3_D_6
    aux_m12_p6 = aux_m12+ M_SQRT3_D_6
    aux_m12_p2 = aux_m12+ M_SQRT3_D_2
    aux_m12_m2 = aux_m12- M_SQRT3_D_2

    aux_mm12 = x1 - x_p0_p12_scaled
    aux_mm12_p6 = aux_mm12 + M_SQRT3_D_6
    aux_mm12_m6 = aux_mm12 - M_SQRT3_D_6
    aux_mm12_m2 = aux_mm12 - M_SQRT3_D_2
    aux_mm12_p2 = aux_mm12 + M_SQRT3_D_2

    aux_pm12 = x1 + x_p0_p12_scaled
    aux_pm12_m6 = aux_pm12 - M_SQRT3_D_6
    aux_pm12_p6 = aux_pm12 + M_SQRT3_D_6
    aux_pm12_p2 = aux_pm12 + M_SQRT3_D_2
    aux_pm12_m2 = aux_pm12 - M_SQRT3_D_2

    aux_m0 = x1 - x_p0_scaled
    aux_m0_p23 = aux_m0 + M_SQRT3_2D3
    aux_m0_m23 = aux_m0 - M_SQRT3_2D3
    aux_m0_p3 = aux_m0 + M_SQRT3_D_3
    aux_m0_m3 = aux_m0 - M_SQRT3_D_3

    aux_p0 = x1 + x_p0_scaled
    aux_p0_m23 = aux_p0 - M_SQRT3_2D3
    aux_p0_p23 = aux_p0 + M_SQRT3_2D3
    aux_p0_p3 = aux_p0 + M_SQRT3_D_3
    aux_p0_m3 = aux_p0 - M_SQRT3_D_3

    h_p0 = Heaviside(x2, ref)
    h_p0_m1 = Heaviside(x_p0_m1, ref)
    h_p0_p1 = Heaviside(x_p0_p1, ref)
    h_p0_m12 = Heaviside(x_p0_m12, ref)
    h_p0_p12 = Heaviside(x_p0_p12, ref)

    h_m0 = Heaviside(-x2, ref)
    h_m0_p12 = Heaviside(1 / 2 - x2, ref)
    h_m0_m12 = Heaviside(-x2 - 1 / 2, ref)
    h_m0_p1 = Heaviside(1 - x2, ref)
    h_m0_m1 = Heaviside(-x2 - 1, ref)

    h_aux_m0 = Heaviside(aux_m0, ref)
    h_aux_m1 = Heaviside(aux_m1, ref)
    h_aux_p0 = Heaviside(aux_p0, ref)
    h_aux_p1 = Heaviside(aux_p1, ref)
    h_aux_m0_p23 = Heaviside(aux_m0_p23, ref)
    h_aux_m0_m23 = Heaviside(aux_m0_m23, ref)
    h_aux_p0_m23 = Heaviside(aux_p0_m23, ref)
    h_aux_p0_p23 = Heaviside(aux_p0_p23, ref)
    h_aux_m1_p3 = Heaviside(aux_m1_p3, ref)
    h_aux_p1_m3 = Heaviside(aux_p1_m3, ref)
    h_aux_m12_m6 = Heaviside(aux_m12_m6, ref)
    h_aux_m12_p2 = Heaviside(aux_m12_p2, ref)
    h_aux_p12_m2 = Heaviside(aux_p12_m2, ref)
    h_aux_p12_p6 = Heaviside(aux_p12_p6, ref)
    h_aux_mm12_m2 = Heaviside(aux_mm12_m2, ref)
    h_aux_mm12_p6 = Heaviside(aux_mm12_p6, ref)
    h_aux_pm12_m6 = Heaviside(aux_pm12_m6, ref)
    h_aux_pm12_p2 = Heaviside(aux_pm12_p2, ref)
    h_aux_mm1_m3 = Heaviside(aux_mm1_m3, ref)
    h_aux_pm1_p3 = Heaviside(aux_pm1_p3, ref)
    h_aux_mm1 = Heaviside(aux_mm1, ref)

    scale = M_SQRT3_2D3
    res = 4 * x2 * aux_m0 * h_p0 * h_aux_m0 - \
          4 * x2 * aux_p0 * h_m0 * h_aux_p0 + \
          x2 * aux_m0_m23 * h_p0 * h_aux_m0_m23 + \
          x2 * aux_m0_p23 * h_p0 * h_aux_m0_p23 - \
          x2 * aux_p0_m23 * h_m0 * h_aux_p0_m23 - \
          x2 * aux_p0_p23 * h_m0 * h_aux_p0_p23 + \
          x_p0_m1 * aux_m1_p3 * h_p0_m1 * h_aux_m1_p3 - \
          x_p0_m1 * aux_p1_m3 * h_m0_p1 * h_aux_p1_m3 - \
          2 * (x_p0_m12) * aux_m12_m6 * h_p0_m12 * h_aux_m12_m6 - \
          2 * (x_p0_m12) * aux_m12_p2 * h_p0_m12 * h_aux_m12_p2 + \
          2 * (x_p0_m12) * aux_p12_m2 * h_m0_p12 * h_aux_p12_m2 + \
          2 * (x_p0_m12) * aux_p12_p6 * h_m0_p12 * h_aux_p12_p6 - \
          2 * (x_p0_p12) * aux_mm12_m2 * h_p0_p12 * h_aux_mm12_m2 - \
          2 * (x_p0_p12) * aux_mm12_p6 * h_p0_p12 * h_aux_mm12_p6 + \
          2 * (x_p0_p12) * aux_pm12_m6 * h_m0_m12 * h_aux_pm12_m6 + \
          2 * (x_p0_p12) * aux_pm12_p2 * h_m0_m12 * h_aux_pm12_p2 + \
          (x_p0_p1) * (aux_mm1_m3) * h_p0_p1 * h_aux_mm1_m3 - \
          (x_p0_p1) * (aux_pm1_p3) * h_m0_m1 * h_aux_pm1_p3 + \
          M_SQRT3 * ((aux_m0) ** 2 * h_p0 * h_aux_m0 +
                     (aux_p0) ** 2 * h_m0 * h_aux_p0 +
                     (aux_m1) ** 2 * h_aux_m1 * h_p0_m1 / 2 +
                     (aux_p1) ** 2 * h_m0_p1 * h_aux_p1 / 2 +
                     (aux_mm1) ** 2 * h_aux_mm1 * h_p0_p1 / 2 +
                     (aux_mm1) ** 2 * h_aux_mm1 * h_m0_m1 / 2 +
                     (aux_m0_m23) ** 2 * h_p0 * h_aux_m0_m23 / 2 +
                     (aux_m0_m3) ** 2 * h_p0 * Heaviside(aux_m0_m3, ref) / 2 +
                     (aux_m0_p3) ** 2 * h_p0 * Heaviside(aux_m0_p3, ref) / 2 +
                     (aux_m0_p23) ** 2 * h_p0 * h_aux_m0_p23 / 2 +
                     (aux_p0_m23) ** 2 * h_m0 * h_aux_p0_m23 / 2 +
                     (aux_p0_m3) ** 2 * h_m0 * Heaviside(aux_p0_m3, ref) / 2 +
                     (aux_p0_p3) ** 2 * h_m0 * Heaviside(aux_p0_p3, ref) / 2 +
                     (aux_p0_p23) ** 2 * h_m0 * h_aux_p0_p23 / 2 -
                     (aux_m12_m2) ** 2 * h_p0_m12 * Heaviside(aux_m12_m2, ref) / 2 -
                     (aux_m12_m6) ** 2 * h_p0_m12 * h_aux_m12_m6 / 2 -
                     (aux_m12_p6) ** 2 * h_p0_m12 * Heaviside(aux_m12_p6, ref) / 2 -
                     (aux_m12_p2) ** 2 * h_p0_m12 * h_aux_m12_p2 / 2 -
                     (aux_p12_m2) ** 2 * h_m0_p12 * h_aux_p12_m2 / 2 -
                     (aux_p12_m6) ** 2 * h_m0_p12 * Heaviside(aux_p12_m6, ref) / 2 -
                     (aux_p12_p6) ** 2 * h_m0_p12 * h_aux_p12_p6 / 2 -
                     (aux_p12_p2) ** 2 * h_m0_p12 * Heaviside(aux_p12_p2, ref) / 2 -
                     (aux_mm12_m2) ** 2 * h_p0_p12 * h_aux_mm12_m2 / 2 -
                     (aux_mm12_m6) ** 2 * h_p0_p12 * Heaviside(aux_mm12_m6, ref) / 2 -
                     (aux_mm12_p6) ** 2 * h_p0_p12 * h_aux_mm12_p6 / 2 -
                     (aux_mm12_p2) ** 2 * h_p0_p12 * Heaviside(aux_mm12_p2, ref) / 2 -
                     (aux_pm12_m2) ** 2 * h_m0_m12 * Heaviside(aux_pm12_m2, ref) / 2 -
                     (aux_pm12_m6) ** 2 * h_m0_m12 * h_aux_pm12_m6 / 2 -
                     (aux_pm12_p6) ** 2 * h_m0_m12 * Heaviside(aux_pm12_p6, ref) / 2 -
                     (aux_pm12_p2) ** 2 * h_m0_m12 * h_aux_pm12_p2 / 2)
    res = res * scale
    return res


def _hexspline2_t1(x1, x2):
    """Hexspline of order p=2 in region t1"""

    # This function has been created with sympy

    x1 = np.asanyarray(np.abs(x1))
    x2 = np.asanyarray(np.abs(x2))

    M_SQRT3_D_3 = M_SQRT3 / 3
    M_SQRT3_D_2 = M_SQRT3 / 2
    M_SQRT3_D_6 = M_SQRT3 / 6
    M_SQRT3_2D3 = 2 * M_SQRT3_D_3

    x_p0_m12 = x2 - 1 / 2
    x_p0_p12 = x2 + 1 / 2

    x_p0_scaled = M_SQRT3_D_3 * x2

    x_p0_m12_scaled = M_SQRT3_D_3 * x_p0_m12
    x_p0_p12_scaled = M_SQRT3_D_3 * x_p0_p12

    aux_p12 = x1 + x_p0_m12_scaled
    aux_p12_p2 = aux_p12 + M_SQRT3_D_2
    aux_p12_p6 = aux_p12 + M_SQRT3_D_6
    aux_mm12 = x1 - x_p0_p12_scaled
    aux_mm12_p2 = aux_mm12 + M_SQRT3_D_2
    aux_m0 = x1 - x_p0_scaled
    aux_m0_p23 = aux_m0 + M_SQRT3_2D3
    aux_m0_p3 = aux_m0 + M_SQRT3_D_3

    scale = M_SQRT3_2D3
    res = x2 * aux_m0_p23  + \
          2 * x_p0_m12 * aux_p12_p6  + \
          M_SQRT3 * (
                     (aux_m0_p3) ** 2 / 2 +
                     (aux_m0_p23) ** 2 / 2 -
                     (aux_p12_p6) ** 2 / 2 -
                     (aux_p12_p2) ** 2  / 2 -
                     (aux_mm12_p2) ** 2 / 2
                  )
    res = res * scale
    return res


def _hexspline2_regions(x1, x2):
    """Hexspline regions of order p=2"""
    x1 = np.asanyarray(x1)
    x2 = np.asanyarray(x2)
    Heaviside = np.heaviside

    M_SQRT3_D_3 = M_SQRT3 / 3
    M_SQRT3_D_2 = M_SQRT3 / 2
    M_SQRT3_D_6 = M_SQRT3 / 6
    M_SQRT3_2D3 = 2 * M_SQRT3_D_3
    ref = 1 / 2

    x_p0_m1 = x2 - 1
    x_p0_p1 = x2 + 1
    x_p0_m12 = x2 - 1 / 2
    x_p0_p12 = x2 + 1 / 2

    x_p0_scaled = M_SQRT3_D_3 * x2
    x_p0_m1_scaled = M_SQRT3_D_3 * x_p0_m1
    x_p0_p1_scaled = M_SQRT3_D_3 * x_p0_p1
    x_p0_m12_scaled = M_SQRT3_D_3 * x_p0_m12
    x_p0_p12_scaled = M_SQRT3_D_3 * x_p0_p12

    aux_p1 = x1 + x_p0_m1_scaled
    aux_p1_m3 = aux_p1 - M_SQRT3_D_3

    aux_m1 = x1 - x_p0_m1_scaled
    aux_m1_p3 = x1 - x_p0_m1_scaled + M_SQRT3_D_3

    aux_mm1 = x1 - x_p0_p1_scaled
    aux_mm1_m3 = aux_mm1 - M_SQRT3_D_3
    aux_pm1 = x1 + M_SQRT3 * (x_p0_p1) / 3
    aux_pm1_p3 = aux_pm1 + M_SQRT3_D_3

    aux_p12 = x1 + x_p0_m12_scaled
    aux_p12_m2 = aux_p12 - M_SQRT3_D_2
    aux_p12_p2 = aux_p12 + M_SQRT3_D_2
    aux_p12_p6 = aux_p12 + M_SQRT3_D_6
    aux_p12_m6 = aux_p12 - M_SQRT3_D_6

    aux_m12 = x1 - x_p0_m12_scaled
    aux_m12_m6 = aux_m12- M_SQRT3_D_6
    aux_m12_p6 = aux_m12+ M_SQRT3_D_6
    aux_m12_p2 = aux_m12+ M_SQRT3_D_2
    aux_m12_m2 = aux_m12- M_SQRT3_D_2

    aux_mm12 = x1 - x_p0_p12_scaled
    aux_mm12_p6 = aux_mm12 + M_SQRT3_D_6
    aux_mm12_m6 = aux_mm12 - M_SQRT3_D_6
    aux_mm12_m2 = aux_mm12 - M_SQRT3_D_2
    aux_mm12_p2 = aux_mm12 + M_SQRT3_D_2

    aux_pm12 = x1 + x_p0_p12_scaled
    aux_pm12_m6 = aux_pm12 - M_SQRT3_D_6
    aux_pm12_p6 = aux_pm12 + M_SQRT3_D_6
    aux_pm12_p2 = aux_pm12 + M_SQRT3_D_2
    aux_pm12_m2 = aux_pm12 - M_SQRT3_D_2

    aux_m0 = x1 - x_p0_scaled
    aux_m0_p23 = aux_m0 + M_SQRT3_2D3
    aux_m0_m23 = aux_m0 - M_SQRT3_2D3
    aux_m0_p3 = aux_m0 + M_SQRT3_D_3
    aux_m0_m3 = aux_m0 - M_SQRT3_D_3

    aux_p0 = x1 + x_p0_scaled
    aux_p0_m23 = aux_p0 - M_SQRT3_2D3
    aux_p0_p23 = aux_p0 + M_SQRT3_2D3
    aux_p0_p3 = aux_p0 + M_SQRT3_D_3
    aux_p0_m3 = aux_p0 - M_SQRT3_D_3

    r = {}
    h_p0 = Heaviside(x2, ref)
    h_p0_m1 = Heaviside(x_p0_m1, ref)
    h_p0_p1 = Heaviside(x_p0_p1, ref)
    h_p0_m12 = Heaviside(x_p0_m12, ref)
    h_p0_p12 = Heaviside(x_p0_p12, ref)

    h_m0 = Heaviside(-x2, ref)
    h_m0_p12 = Heaviside(1 / 2 - x2, ref)
    h_m0_m12 = Heaviside(-x2 - 1 / 2, ref)
    h_m0_p1 = Heaviside(1 - x2, ref)
    h_m0_m1 = Heaviside(-x2 - 1, ref)

    h_aux_m0 = Heaviside(aux_m0, ref)
    h_aux_m1 = Heaviside(aux_m1, ref)
    h_aux_p0 = Heaviside(aux_p0, ref)
    h_aux_p1 = Heaviside(aux_p1, ref)
    h_aux_m0_p23 = Heaviside(aux_m0_p23, ref)
    h_aux_m0_m23 = Heaviside(aux_m0_m23, ref)
    h_aux_p0_m23 = Heaviside(aux_p0_m23, ref)
    h_aux_p0_p23 = Heaviside(aux_p0_p23, ref)
    h_aux_m1_p3 = Heaviside(aux_m1_p3, ref)
    h_aux_p1_m3 = Heaviside(aux_p1_m3, ref)
    h_aux_m12_m6 = Heaviside(aux_m12_m6, ref)
    h_aux_m12_p2 = Heaviside(aux_m12_p2, ref)
    h_aux_p12_m2 = Heaviside(aux_p12_m2, ref)
    h_aux_p12_p6 = Heaviside(aux_p12_p6, ref)
    h_aux_mm12_m2 = Heaviside(aux_mm12_m2, ref)
    h_aux_mm12_p6 = Heaviside(aux_mm12_p6, ref)
    h_aux_pm12_m6 = Heaviside(aux_pm12_m6, ref)
    h_aux_pm12_p2 = Heaviside(aux_pm12_p2, ref)
    h_aux_mm1_m3 = Heaviside(aux_mm1_m3, ref)
    h_aux_pm1_p3 = Heaviside(aux_pm1_p3, ref)
    h_aux_mm1 = Heaviside(aux_mm1, ref)

    h_aux_m0_m3 = Heaviside(aux_m0_m3, ref)
    h_aux_m0_p3 = Heaviside(aux_m0_p3, ref)
    h_aux_p0_m3 = Heaviside(aux_p0_m3, ref)
    h_aux_p0_p3 = Heaviside(aux_p0_p3, ref)

    h_aux_m12_m2 = Heaviside(aux_m12_m2, ref)
    h_aux_m12_p6 = Heaviside(aux_m12_p6, ref)
    h_aux_p12_m6 = Heaviside(aux_p12_m6, ref)
    h_aux_p12_p2 = Heaviside(aux_p12_p2, ref)
    h_aux_mm12_m6 = Heaviside(aux_mm12_m6, ref)
    h_aux_mm12_p2 = Heaviside(aux_mm12_p2, ref)
    h_aux_pm12_m2 = Heaviside(aux_pm12_m2, ref)
    h_aux_pm12_p6 = Heaviside(aux_pm12_p6, ref)

    inter = locals().copy()
    for key in inter:
        if key.startswith('h_'):
            r[key] = inter[key]
    return r


def hexspline_support(xx, yy, p):
    """Support of hexspline of order p

    The support is an hexagon of area Omega p^2
    Omega = M_SQRT3 / 2
    """
    return hex_c(xx, yy, rad=p / M_SQRT3, ang=0.0)


def hexspline_bbox(p):
    """Bounding box of hexspline of order p

    The support is an hexagon of area Omega p^2
    Omega = M_SQRT3 / 2
    """
    # (x1, x2), (y1, y2)
    # TODO: review numbers
    x1 = p / M_SQRT3
    y1 = p / 2.0
    return (-x1, x1), (-y1, y1)


def hexspline_gauss(xx, yy, p):
    """Approximation of the p-order hexspline by a bivariate normal

    https://miplab.epfl.ch/pub/vandeville0202.pdf

    Least-squares spline resampling to a hexagonal lattice
    Signal Processing:Image Communication17 (2002) 393-408

    Eq B.4
    """
    from scipy.stats import multivariate_normal
    zz = np.array([xx, yy])
    zz = np.moveaxis(zz, 0, -1)
    in_support = hexspline_support(xx, yy, p)

    coeff = 5 * M_SQRT3 / 144
    omega = M_SQRT3 / 2
    sigma = coeff * np.eye(2)
    covar = p * sigma / omega

    out = omega * multivariate_normal.pdf(zz, mean=[0, 0], cov=covar)
    out[~in_support] = 0
    return out


def OsincH_2pi(x, y):
    t1 = np.cos(-x / (2 * M_SQRT3) + y / 2) - np.cos(x / M_SQRT3)
    t2 = np.cos(x / (2 * M_SQRT3) + y / 2) - np.cos(x / M_SQRT3)
    ts = t1 / (x + M_SQRT3 * y) + t2 / (x - M_SQRT3 * y)
    return 2 * M_SQRT3 * ts / x


def sincH(x, y):
    """Generalization of the sinc function to a hexagonal grid

    The function is 1 in (0,0) and 0 in all the knots of the grid
    """
    # This expression seems numerically unstable near 0
    # it should be 1
    x = np.asanyarray(x)
    y = np.asanyarray(y)

    xx = 2 * np.pi * np.where(x == 0, 1.0e-20, x)
    yy = 2 * np.pi * np.where(y == 0, 1.0e-20, y)

    t3 = np.cos(xx / M_SQRT3)
    t1 = np.cos(-xx / (2 * M_SQRT3) + yy / 2.0) - t3
    t2 = np.cos(xx / (2 * M_SQRT3) + yy / 2.0) - t3
    ts = t1 / (xx + M_SQRT3 * yy) + t2 / (xx - M_SQRT3 * yy)
    return 4 * ts / xx


# Convolution, compute kernel
def rescaling_kernel(p, scale=1):
    """Rescaling kernel from hexgrid to rectangular grid"""
    from megaradrp.simulation.convolution import setup_grid
    from scipy import signal
    from scipy.interpolate import RectBivariateSpline

    Dx = 0.005
    Dy = 0.005
    DA = Dx * Dy
    # TODO: the support should be computed from the scale and p
    xsize = ysize = 3.0
    xx, yy, xs, ys, xl, yl = setup_grid(xsize, ysize, Dx, Dy)

    detR0 = M_SQRT3 / 2
    detR1 = scale * scale

    # index of bspline
    n = p - 1

    rect_kernel = signal.bspline(xx / scale, n) * signal.bspline(yy / scale, n) / detR1

    hex1 = hexspline1(xx, yy)

    if p == 1:
        hex_kernel = hex1
    elif p == 2:
        hex2 = signal.fftconvolve(hex1, hex1, mode='same') * DA / detR0
        hex_kernel = hex2
    elif p == 3:
        hex2 = signal.fftconvolve(hex1, hex1, mode='same') * DA / detR0
        hex3 = signal.fftconvolve(hex2, hex1, mode='same') * DA / detR0
        hex_kernel = hex3
    else:
        raise ValueError('p>3 not implemented')

    kernel = signal.fftconvolve(rect_kernel, hex_kernel, mode='same') * DA
    rbs = RectBivariateSpline(xs, ys, kernel)
    return rbs



#
# Copyright 2019-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime

from astropy.io import fits
import numpy
from numina.processing import Corrector


_logger = logging.getLogger(__name__)


class DiffuseLightCorrector(Corrector):
    """A Node that corrects a frame from diffuse light"""

    def __init__(self, diffuse, datamodel=None, calibid='calibid-unknown',
                 dtype='float32'):

        super(DiffuseLightCorrector, self).__init__(
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype)

        if isinstance(diffuse, fits.HDUList):
            self.corr = diffuse[0].data
        elif isinstance(diffuse, fits.ImageHDU):
            self.corr = diffuse.data
        else:
            self.corr = numpy.asarray(diffuse)

    def header_update(self, hdr, imgid):
        hdr['NUM-DFL'] = self.calibid
        hdr['history'] = f'Diffuse light correction {imgid}'
        hdr['history'] = f'Diffuse light correction time {datetime.datetime.utcnow().isoformat()}'

    def run(self, img):
        imgid = self.get_imgid(img)
        _logger.debug('correct diffuse light in image %s', imgid)

        img['primary'].data -= self.corr
        hdr = img['primary'].header

        self.header_update(hdr, imgid)

        return img



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Load image correctors according to present calibrations"""


import logging


import numina.util.node as node
import numina.processing as proc

from megaradrp.processing.trimover import OverscanCorrector, TrimImage
from megaradrp.processing.trimover import GainCorrector
from megaradrp.processing.slitflat import SlitFlatCorrector
from megaradrp.processing.diffuselight import DiffuseLightCorrector


_logger = logging.getLogger(__name__)


def get_corrector_bpm(rinput, meta, ins, datamodel):
    bpm_info = meta.get('master_bpm')
    if bpm_info is not None:
        with rinput.master_bpm.open() as hdul:
            _logger.info('loading BPM')
            mbpm = hdul[0].data
            calibid = datamodel.get_imgid(hdul)
            _logger.debug('BPM image: %s', calibid)
            bpm_corrector = proc.BadPixelCorrector(
                mbpm,
                datamodel=datamodel,
                calibid=calibid,
                hwin=0,
                wwin=3
            )
    else:
        _logger.info('BPM not provided, ignored')
        bpm_corrector = node.IdNode()

    return bpm_corrector


def get_corrector_super(rinput, meta, key, correctorclass, datamodel):
    info = meta.get(key)
    req = getattr(rinput, key)
    if req is not None:
        with req.open() as hdul:
            _logger.info('loading %s', key)
            _logger.debug('%s info: %s', key, info)
            datac = hdul['primary'].data
            calibid = datamodel.get_imgid(hdul)
            corrector = correctorclass(datac, datamodel=datamodel,
                                       calibid=calibid)
    else:
        _logger.info('%s not provided, ignored', key)
        corrector = node.IdNode()

    return corrector


def get_corrector_bias(rinput, meta, ins, datamodel):
    key = 'master_bias'
    correctorclass = proc.BiasCorrector
    return get_corrector_super(rinput, meta, key, correctorclass, datamodel)


def get_corrector_dark(rinput, meta, ins, datamodel):
    key = 'master_dark'
    correctorclass = proc.DarkCorrector
    return get_corrector_super(rinput, meta, key, correctorclass, datamodel)


def get_corrector_slit_flat(rinput, meta, ins, datamodel):
    key = 'master_slitflat'
    info = meta.get(key)
    if info is not None:
        req = getattr(rinput, key)
        with req.open() as hdul:
            _logger.info('loading slit flat')
            _logger.debug('%s image: %s', key, info)
            mbpm = hdul[0].data
            calibid = datamodel.get_imgid(hdul)
            corrector = SlitFlatCorrector(mbpm, datamodel, calibid=calibid)
    else:
        _logger.info('%s not provided, ignored', key)
        corrector = node.IdNode()

    return corrector


def get_corrector_diffuse_light(rinput, meta, ins, datamodel):
    key = 'diffuse_light_image'
    info = meta.get(key)
    if info is not None:
        req = getattr(rinput, key)
        with req.open() as hdul:
            _logger.info('loading diffuse light image')
            _logger.debug('%s image: %s', key, info)
            mbpm = hdul[0].data
            calibid = datamodel.get_imgid(hdul)
            corrector = DiffuseLightCorrector(mbpm, datamodel, calibid=calibid)
    else:
        _logger.info('%s not provided, ignored', key)
        corrector = node.IdNode()

    return corrector


def get_corrector_overscan(rinput, meta, ins, datamodel):
    detconf = ins.get_property('detector.scan')
    return OverscanCorrector(
        detconf,
        datamodel=datamodel,
        calibid=str(ins.get_device('detector').origin.uuid)
    )


def get_corrector_trimming(rinput, meta, ins, datamodel):
    detconf = ins.get_property('detector.scan')
    return TrimImage(
        detconf,
        datamodel=datamodel,
        calibid=str(ins.get_device('detector').origin.uuid)
    )


def get_corrector_gain(rinput, meta, ins, datamodel):
    """Correct from gain"""
    return GainCorrector(
        detconf=ins.get_property('detector.scan'),
        datamodel=datamodel,
        calibid=str(ins.get_device('detector').origin.uuid)
    )



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""


class GeometricAperture(object):
    def __init__(self, fibid, boxid, start, stop):
        self.fibid = fibid
        self.boxid = boxid
        self.start = start
        self.stop = stop

    @property
    def valid(self):
        return self.is_valid()

    def aper_center(self):
        raise NotImplementedError

    def is_valid(self):
        return True

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__ = state



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Compute centroids in MEGARA RSS images """

import math
import logging

import numpy
from scipy.spatial import KDTree
from numina.constants import FWHM_G

from megaradrp.instrument.focalplane import FocalPlaneConf


_logger = logging.getLogger(__name__)


def find_brightest_spaxel(final, extraction_region):
    """Find the centroid around the brightest spaxel"""
    fp_conf = FocalPlaneConf.from_img(final)
    _logger.debug("LCB configuration is %s", fp_conf.conf_id)

    rssdata = final[0].data
    cut1, cut2 = extraction_region

    flux_per_cell_all = rssdata[:, cut1:cut2].mean(axis=1)

    max_cell = flux_per_cell_all.argmax() + 1
    max_fiber_ = fp_conf.fibers[max_cell]

    _logger.info("maximum flux in spaxel %d -- %s", max_cell, max_fiber_.name)
    return max_fiber_


def calc_centroid(final, extraction_region, point, nrings):
    """compute the centroid around point"""
    import megaradrp.instrument.constants as cons
    import megaradrp.datamodel as dm

    fp_conf = FocalPlaneConf.from_img(final)
    _logger.debug("LCB configuration is %s", fp_conf.conf_id)

    rssdata = final[0].data
    cut1, cut2 = extraction_region

    points = [point]

    fibers = fp_conf.connected_fibers(valid_only=True)

    grid_coords = []
    for fiber in fibers:
        grid_coords.append((fiber.x, fiber.y))
    # setup kdtree for searching
    kdtree = KDTree(grid_coords)

    # Other possibility is
    # query using radius instead
    # radius = 1.2
    # kdtree.query_ball_point(points, k=7, r=radius)
    _logger.debug('adding %d nrings', nrings)
    npoints = 1 + 3 * nrings * (nrings + 1)
    _logger.debug('adding %d fibers', npoints)

    dis_p, idx_p = kdtree.query(points, k=npoints)

    _logger.info('Using %d nearest fibers', npoints)

    scale, funit = dm.fiber_scale_unit(final, unit=True)
    _logger.debug('unit is %s', funit)

    platescale = cons.GTC_FC_A_PLATESCALE.value
    positions = []
    for diss, idxs, point in zip(dis_p, idx_p, points):
        # For each point
        value = [p * scale for p in point]
        value_mm = [(v / platescale) for v in value]
        _logger.info('For point %s arcsec', value)
        _logger.info('For point %s mm', value_mm)
        colids = []
        coords = []
        for dis, idx in zip(diss, idxs):
            fiber = fibers[idx]
            colids.append(fiber.fibid - 1)
            coords.append((fiber.x, fiber.y))
        _logger.debug('nearest fibers')
        _logger.debug('%s', [col + 1 for col in colids])
        coords = numpy.asarray(coords) * scale
        # flux_per_cell = flux_per_cell_all[colids]
        flux_per_cell = rssdata[colids, cut1:cut2].mean(axis=1)
        flux_per_cell_total = flux_per_cell.sum()
        flux_per_cell_norm = flux_per_cell / flux_per_cell_total
        # centroid
        scf = coords.T * flux_per_cell_norm
        centroid = scf.sum(axis=1)
        _logger.info('centroid: %s arcsec', list(centroid))
        _logger.info('centroid: %s mm', list(centroid / platescale))
        # central coords
        c_coords = coords - centroid
        scf0 = scf - centroid[:, numpy.newaxis] * flux_per_cell_norm
        mc2 = numpy.dot(scf0, c_coords)
        _logger.info('2nd order moments, x2=%f, y2=%f, xy=%f arcsec^2', mc2[0, 0], mc2[1, 1], mc2[0, 1])
        if (mc2[0, 0] > 0) and (mc2[1, 1] > 0):
            _logger.info('FWHM , x=%f, y=%f arcsec',
                             FWHM_G * math.sqrt(mc2[0, 0]),
                             FWHM_G * math.sqrt(mc2[1, 1])
                             )
        positions.append(centroid / platescale)
    return positions[0]


def calc_centroid_brightest(final, extraction_region, nrings):
    """Compute the centroid around the brightest spaxel"""
    max_fiber_ = find_brightest_spaxel(final, extraction_region)
    point = (max_fiber_.x, max_fiber_.y)
    # The brightest spaxel
    position = calc_centroid(final, extraction_region, point, nrings)
    return position


#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Fix RSS imperfections
"""
import math

import numpy as np

from megaradrp.instrument.focalplane import FocalPlaneConf
import megaradrp.processing.wcs as mwcs


def fix_missing_fiber(rss, fibid):
    """Interpolate missing fiber fibid"""
    hdr = rss['FIBERS'].header
    fp = FocalPlaneConf.from_img(rss)
    # Fibers around 623 are
    idxs = fp.nearby_fibers(fibid)
    avg = np.zeros_like(rss[0].data[fibid - 1])
    for idx in idxs:
        avg += rss[0].data[idx - 1]
    avg /= len(idxs)

    l1fmt = "FIB{:03d}W1"
    l2fmt = "FIB{:03d}W2"
    try:
        # Change limits in header to the min-max of surrounding fibers
        for sub, func in [(l1fmt, max), (l2fmt, min)]:
            keys = []
            for idx in idxs:
                keys.append(hdr[sub.format(idx)])
            hdr[sub.format(fibid)] = func(keys)

        l1 = hdr[l1fmt.format(fibid)]
        l2 = hdr[l2fmt.format(fibid)]
        if l1 > 1:
            avg[0:l1 - 1] = 0
        if l2 < len(avg):
            avg[l2 - 1:] = 0
    except KeyError:
        pass

    rss[0].data[fibid - 1] = avg
    return rss


def recompute_wcs(hdr, ipa):
    """Recompute the WCS rotations from IPA """
    pa = mwcs.compute_pa_from_ipa(ipa)
    # print('IPA angle is:', ipa, 'PA angle is', math.fmod(pa, 360))
    x = hdr['PC1_1']
    y = hdr['PC1_2']
    # print('PA from header is:', np.rad2deg(math.atan2(y, x)))
    return mwcs.update_wcs_from_ipa(hdr, pa)




import numpy
import pytest

from ..utils import atleast_2d_last

@pytest.mark.parametrize("arr, shape, ndim", [
    ([1, 2, 3], (3, 1), 2),
    ([1, 2, 3, 789], (4, 1), 2),
    (4, (1, 1), 2),
    ([[1,2,3], [4,5,6]], (2,3), 2),
    (numpy.empty((3,3,3)), (3,3,3), 3)
])
def test_utils_atleast1(arr, shape, ndim):
    # With one input, return de object
    res = atleast_2d_last(arr)
    print(res, res.shape)
    assert res.shape == shape
    assert res.ndim == ndim

@pytest.mark.parametrize("arrs, shapes, ndims", [
    ([[1, 2, 3], [1, 2, 3, 789], 4, [[1, 2, 3], [4, 5, 6]], numpy.empty((3, 3, 3))],
     [(3, 1), (4, 1), (1, 1), (2, 3), (3, 3, 3)],
     [2, 2, 2, 2, 3])
])
def test_utils_atleast2(arrs, shapes, ndims):
    # With more than one, a list
    res = atleast_2d_last(*arrs)
    assert isinstance(res, list)
    for el, shape, ndim in zip(res, shapes, ndims):
        assert el.shape == shape
        assert el.ndim == ndim




import enum


class DataOrigin(enum.Enum):
    UNKNOWN = 0
    OBSERVED = 1
    PROCESSED = 2
    GENERATED = 3


class MegaraDataType(enum.Enum):
    UNKNOWN = 1
    IMAGE_RAW = 100
    IMAGE_BIAS = 102
    IMAGE_DARK = 103
    IMAGE_SLITFLAT = 104
    IMAGE_FLAT = 105
    IMAGE_COMP = 106
    #
    IMAGE_TWILIGHT = 107
    IMAGE_TEST = 109
    IMAGE_TARGET = 150
    #
    IMAGE_PROCESSED = 200
    MASTER_BPM = 201
    MASTER_BIAS = 202
    MasterBias = 202 # Alias
    MASTER_DARK = 203
    MASTER_SLITFLAT = 204
    DIFFUSE_LIGHT = 211
    #
    RSS_PROCESSED = 300
    MASTER_FLAT = 305
    MasterFiberFlat = 305 # Alias
    MASTER_TWILIGHT = 306
    RSS_WL_PROCESSED = 400
    SPEC_PROCESSED = 500
    MASTER_SENSITIVITY = 503
    STRUCT_PROCESSED = 600
    TRACE_MAP = 601
    MODEL_MAP = 602
    WAVE_CALIB = 603


#
# Copyright 2017-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Enumerated types for instrument configurations of MEGARA"""

import enum


class VphName(enum.Enum):
    VPH405_LR = 1405
    VPH480_LR = 1480
    VPH570_LR = 1570
    VPH675_LR = 1675
    VPH799_LR = 1799
    VPH890_LR = 1890
    VPH410_MR = 2410
    VPH443_MR = 2443
    VPH481_MR = 2481
    VPH521_MR = 2521
    VPH567_MR = 2567
    VPH617_MR = 2617
    VPH656_MR = 2656
    VPH712_MR = 2712
    VPH777_MR = 2777
    VPH926_MR = 2926
    VPH665_HR = 3665
    VPH863_HR = 3863


class BundleType(enum.Enum):
    """Types of bundles"""
    LCB = 1
    RP = 2
    SKY = 3


class FiberPatternType(enum.Enum):
    """Types of fiber pattern on bundles"""
    FIXED = 1
    RP = 2
    SKY = 3


class SlitType(enum.Enum):
    """Types of slits"""
    LCB = 1
    MOS = 2


class SlitPosition(enum.Enum):
    """Positions of the pseudo slit"""
    LCB = 1
    MOS = 2
    OPEN = 3


class TargetType(enum.Enum):
    """Possible targets in a fiber bundle"""
    SOURCE = 1
    UNKNOWN = 2
    UNASSIGNED = 3
    SKY = 4
    REFERENCE = 5
    # aliases for the other fields
    STAR = 5
    BLANK = 4


# Related with Robotic positioner sequences

FIBERMOS_MOV_MAXR1 = 194086
FIBERMOS_MOV_MAXR2 = 0
FIBERMOS_MOV_MINR1 = 0
FIBERMOS_MOV_MINR2 = -15360


class RpMovAxis(enum.Enum):
    RP_MOV_AXIS1 = 1
    RP_MOV_AXIS2 = 2


class SequenceType(enum.Enum):
    SEQUENCE_POS_DEPOS = 1
    SEQUENCE_DEPOS_ONLY = 2



#
# Copyright 2015-2021 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the bpm mode recipe module."""

import astropy.io.fits as fits
import numpy as np
from numina.core import DataFrame, ObservationResult
import numina.instrument.assembly as asb


def generate_bias(detector, number, temporary_path):
    from megaradrp.simulation.actions import simulate_bias
    from megaradrp.recipes.calibration.bias import BiasRecipe

    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    date_obs = '2017-11-09T11:00:00.0'
    fs = [simulate_bias(detector) for i in range(number)]
    header = fits.Header()
    header['DATE-OBS'] = date_obs
    header['INSCONF'] = config_uuid
    header['INSTRUME'] = 'MEGARA'
    header['VPH'] = 'LR-U'
    header['INSMODE'] = 'MOS'
    for aux in range(len(fs)):
        fits.writeto(f'{temporary_path}/bias_{aux}.fits', fs[aux],
                     header=header,
                     overwrite=True)

    fs = [f"{temporary_path}/bias_{i}.fits" for i in range(number)]

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'bias_image'

    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)
    insmodel = asb.assembly_instrument(store, config_uuid, date_obs, by_key='uuid')
    insmodel.configure_with_header(header)
    ob.configuration = insmodel
    ob.frames = [DataFrame(filename=f) for f in fs]

    recipe = BiasRecipe()
    ri = recipe.create_input(obresult=ob)
    return recipe.run(ri)


def crear_archivos(temporary_path, number=5):
    from megaradrp.simulation.actions import simulate_flat
    from megaradrp.instrument.components.detector import ReadParams, MegaraDetectorSat
    from megaradrp.recipes.calibration.bpm import BadPixelsMaskRecipe

    config_uuid = '4fd05b24-2ed9-457b-b563-a3c618bb1d4c'
    date_obs = '2017-11-09T11:00:00.0'
    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50
    ron = 2.0
    gain = 1.0
    bias = 1000.0

    qe = 0.8 * np.ones(DSHAPE)
    qe[0:15, 0:170] = 0.0

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN,
                                 qe=qe,
                                 dark=(3.0 / 3600.0),
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    source2 = 1.0

    fs = [simulate_flat(detector, exposure=1.0, source=5000 * source2) for i in
          range(number)]

    header = fits.Header()
    header['DATE-OBS'] = date_obs
    header['INSCONF'] = config_uuid
    header['INSTRUME'] = 'MEGARA'
    header['VPH'] = 'LR-U'
    header['INSMODE'] = 'MOS'
    for aux in range(len(fs)):
        fits.writeto(f'{temporary_path}/flat_{aux}.fits', fs[aux],
                     header=header,
                     overwrite=True)

    result = generate_bias(detector, number, temporary_path)
    result.master_bias.frame.writeto(
        f'{temporary_path}/master_bias_data0.fits',
        overwrite=True)  # Master Bias

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'bias_image'
    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)
    insmodel = asb.assembly_instrument(store, config_uuid, date_obs, by_key='uuid')
    insmodel.configure_with_header(header)
    ob.configuration = insmodel

    names = []
    for aux in range(number):
        names.append(f'{temporary_path}/flat_{aux}.fits')
    ob.frames = [DataFrame(filename=open(nombre).name) for nombre in names]

    recipe = BadPixelsMaskRecipe()
    ri = recipe.create_input(obresult=ob, master_bias=DataFrame(
        filename=open(temporary_path + '/master_bias_data0.fits').name))
    aux = recipe.run(ri)
    aux.master_bpm.frame.writeto(f'{temporary_path}/master_bpm.fits', overwrite=True)

    return names



#
# Copyright 2019-2020 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

import numpy

import megaradrp.instrument.focalplane as fp
from numina.frame.utils import copy_img


def subtract_sky(img, ignored_sky_bundles=None, logger=None):
    # Sky subtraction

    if logger is None:
        logger = logging.getLogger(__name__)

    logger.info('obtain fiber information')
    sky_img = copy_img(img)
    final_img = copy_img(img)
    fp_conf = fp.FocalPlaneConf.from_img(sky_img)
    # Sky fibers
    skyfibs = fp_conf.sky_fibers(valid_only=True,
                                   ignored_bundles=ignored_sky_bundles)
    logger.debug('sky fibers are: %s', skyfibs)
    # Create empty sky_data
    target_data = img[0].data

    target_map = img['WLMAP'].data
    sky_data = numpy.zeros_like(img[0].data)
    sky_map = numpy.zeros_like(img['WLMAP'].data)
    sky_img[0].data = sky_data

    for fibid in skyfibs:
        rowid = fibid - 1
        sky_data[rowid] = target_data[rowid]
        sky_map[rowid] = target_map[rowid]
    # Sum
    coldata = sky_data.sum(axis=0)
    colsum = sky_map.sum(axis=0)

    # Divide only where map is > 0
    mask = colsum > 0
    avg_sky = numpy.zeros_like(coldata)
    avg_sky[mask] = coldata[mask] / colsum[mask]

    # This should be done only on valid fibers
    logger.info('ignoring invalid fibers: %s', fp_conf.invalid_fibers())
    for fibid in fp_conf.valid_fibers():
        rowid = fibid - 1
        final_img[0].data[rowid, mask] = img[0].data[rowid, mask] - avg_sky[mask]
    # Update headers
    #
    return final_img, img, sky_img


def subtract_sky_rss(img, sky_img, ignored_sky_bundles=None, logger=None):
    """Subtract a sky image from an image"""
    # Sky subtraction

    if logger is None:
        logger = logging.getLogger(__name__)

    #logger.info('obtain fiber information')
    final_img = copy_img(img)
    # fiberconf_sky = dm.get_fiberconf(sky_img)
    # fiberconf_target = dm.get_fiberconf(img)

    logger.debug('using WLMAP extension to compute valid regions')

    v_map = img['WLMAP'].data > 0
    sky_map = numpy.zeros_like(img['WLMAP'].data)
    sky_data = sky_img[0].data
    sky_map[:] = v_map[:]

    # This should be done only on valid fibers
    #logger.info('ignoring invalid fibers: %s', fiberconf_target.invalid_fibers())
    final_img[0].data[v_map] = img[0].data[v_map] - sky_data[v_map]
    final_img[0].data[~v_map] = 0.0
    # Update headers
    #
    return final_img, img, sky_img


#
# Copyright 2014-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Load MEGARA DRP"""

from numina.core import drp_load
import numina.core.config as cfg


class MegaraDrpLoader(object):
    """Custom loader class

    This class modifies the rawimage field of the observing modes
    of MEGARA
    """
    @staticmethod
    def mode_loader(mode_node):
        import megaradrp.datamodel as DM
        if 'rawimage' in mode_node:
            rname = mode_node['rawimage']
            mode_node['rawimage'] = DM.MegaraDataType[rname]
        return mode_node


def load_drp():
    """Entry point to load MEGARA DRP."""
    return drp_load('megaradrp', 'drp.yaml', confclass=MegaraDrpLoader)


def is_fits_megara(pathname):
    "Check is any FITS"
    import astropy.io.fits as fits
    # FIXME: incomplete
    if pathname.endswith('.fits') or pathname.endswith('.fits.gz'):
        with fits.open(pathname) as hdulist:
            prim = hdulist[0].header
            instrument = prim.get("INSTRUME", "unknown")
            if instrument == "MEGARA":
                return True
    else:
        return False


@cfg.describe.register('image/fits', is_fits_megara, priority=15)
def describe_fits_megara(pathname):
    import megaradrp.datamodel as DM
    import astropy.io.fits as fits
    with fits.open(pathname) as hdulist:
        return DM.describe_hdulist_megara(hdulist)


@cfg.check.register('MEGARA')
def check_obj_megara(obj, astype=None, level=None):
    import megaradrp.datamodel as DM
    return DM.check_obj_megara(obj, astype=astype, level=level)



from __future__ import division
from __future__ import print_function

import argparse
import numpy as np
from numpy.polynomial import Polynomial
import sys
import numina.instrument.assembly as asb
import numina.types.structured as structured
from numina.array.display.ximshow import ximshow_file
from numina.array.display.pause_debugplot import pause_debugplot


def assign_boxes_to_fibers(pseudo_slit_config, insmode):
    """Read boxes in configuration file and assign values to fibid

    Parameters
    ----------
    pseudo_slit_config : dict
        Contains the association of fibers and boxes
    insmode : string
        Value of the INSMODE keyword: 'LCB' or 'MOS'.

    Returns
    -------
    fibid_with_box : list of strings
        List with string label that contains both the fibid and the
        box name.

    """
    fibid_with_box = []
    n1 = 1
    list_to_print = []
    for dumbox in pseudo_slit_config:
        nfibers = dumbox['nfibers']
        name = dumbox['name']
        n2 = n1 + nfibers
        fibid_with_box += \
            [f"{val1}  [{val2}]"
             for val1, val2 in zip(range(n1, n2), [name] * nfibers)]
        dumstr =f'Box {name:>2},  fibers {n1:3d} - {n2 - 1:3d}'
        list_to_print.append(dumstr)
        n1 = n2
    print(f'\n* Fiber description for INSMODE={insmode}')
    for dumstr in reversed(list_to_print):
        print(dumstr)
    print('---------------------------------')

    return fibid_with_box


def plot_aper(ax, center_model, xmin, xmax, ix_offset,
               rawimage, fibids, fiblabel, colour, correction=0):
    if xmin == xmax == 0:
        num = 4096
        xp = np.linspace(start=1, stop=4096, num=num)
    else:
        num = int(float(xmax - xmin + 1) + 0.5)
        xp = np.linspace(start=xmin, stop=xmax, num=num)
    yp = center_model(xp) + correction
    if rawimage:
        lcut = (yp > 2056.5)
        yp[lcut] += 100
    ax.plot(xp + ix_offset, yp + 1, color=colour, linestyle='dotted')
    if fibids:
        if xmin == xmax == 0:
            xmidpoint = 2048
        else:
            xmidpoint = (xmin+xmax)/2
        ax.text(xmidpoint, yp[int(num / 2)], fiblabel, fontsize=6,
                bbox=dict(boxstyle="round,pad=0.1", fc="white", ec="grey", ),
                color=colour, fontweight='bold', backgroundcolor='white',
                ha='center')


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: overplot traces"
    )
    # positional parameters
    parser.add_argument("fits_file",
                        help="FITS image containing the spectra",
                        type=argparse.FileType('r'))
    parser.add_argument("traces_file",
                        help="JSON file with fiber traces",
                        type=argparse.FileType('r'))
    # optional parameters
    parser.add_argument("--rawimage",
                        help="FITS file is a RAW image (otherwise trimmed "
                             "image is assumed)",
                        action="store_true")
    parser.add_argument("--global_offset", "--global-offset",
                        nargs='+', type=float,
                        help="Global offset polynomial coefficients "
                             "(+upwards, -downwards)")
    parser.add_argument("--fibids",
                        help="Display fiber identification number",
                        action="store_true")
    parser.add_argument("--verbose",
                        help="Enhance verbosity",
                        action="store_true")
    parser.add_argument("--z1z2",
                        help="tuple z1,z2, minmax or None (use zscale)")
    parser.add_argument("--bbox",
                        help="bounding box tuple: nc1,nc2,ns1,ns2")
    parser.add_argument("--keystitle",
                        help="tuple of FITS keywords.format: " +
                             "key1,key2,...keyn.'format'")
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy",
                        default="0,0,640,480")
    parser.add_argument("--pdffile",
                        help="ouput PDF file name",
                        type=argparse.FileType('w'))
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # global_offset in command line
    if args.global_offset is None:
        args_global_offset_set = False
        args.global_offset = [0.0]
    else:
        args_global_offset_set = True

    args_global_offset = Polynomial(args.global_offset)

    # read pdffile
    if args.pdffile is not None:
        from matplotlib.backends.backend_pdf import PdfPages
        pdf = PdfPages(args.pdffile.name)
    else:
        pdf = None

    ax = ximshow_file(args.fits_file.name,
                      args_cbar_orientation='vertical',
                      args_z1z2=args.z1z2,
                      args_bbox=args.bbox,
                      args_keystitle=args.keystitle,
                      args_geometry=args.geometry,
                      pdf=pdf,
                      show=False)

    # trace offsets for RAW images
    if args.rawimage:
        ix_offset = 51
    else:
        ix_offset = 1

    # read and display traces from JSON file
    # TODO: some checks, this should be done by validating the struct
    with open(args.traces_file.name, mode='r') as fd:
        import json
        data = json.load(fd)
        if 'type_fqn' not in data:
            raise ValueError("malformed JSON file, 'type_fqn' missing")
    #
    apers = structured.open(args.traces_file.name)
    # Load metadata from the traces
    meta_info = apers.meta_info

    origin = meta_info['origin']
    insconf_uuid = origin['insconf_uuid']
    # None is allowed
    date_obs = origin.get('date_obs')

    tags = apers.tags
    insmode = tags['insmode']

    # create instrument model
    pkg_paths = ['megaradrp.instrument.configs']
    store = asb.load_paths_store(pkg_paths)

    insmodel = asb.assembly_instrument(store, insconf_uuid, date_obs, by_key='uuid')

    pseudo_slit_config = insmodel.get_value('pseudoslit.boxes', **tags)

    fibid_with_box = list(assign_boxes_to_fibers(pseudo_slit_config, insmode))
    total_fibers = apers.total_fibers
    if total_fibers != len(fibid_with_box):
        raise ValueError('Mismatch between number of fibers and '
                         'expected number from account from boxes')

    if args_global_offset_set:
        global_offset = args_global_offset
    else:
        global_offset = apers.global_offset

    print('>>> Using global_offset:', global_offset)
    ref_column = apers.ref_column

    for geot in apers.contents:
        fibid = geot.fibid
        fiblabel = fibid_with_box[fibid - 1]
        start = geot.start
        stop = geot.stop
        # skip fibers without trace
        if geot.valid:
            center_model = geot.aper_center()
            y_at_ref_column = center_model(ref_column)
            correction = global_offset(y_at_ref_column)
            plot_aper(ax, center_model, start, stop, ix_offset, args.rawimage,
                       args.fibids, fiblabel, colour='blue', correction=correction)
        else:
            print('Warning ---> Missing fiber:', fibid_with_box[fibid - 1])

    if pdf is not None:
        pdf.savefig()
        pdf.close()
    else:
        pause_debugplot(12, pltshow=True, tight_layout=True)


if __name__ == "__main__":

    main()



import pytest
import astropy.io.fits as fits
import numina.instrument.generic
import numina.core
from megaradrp.loader import load_drp


def create_simple_frame():
    hdr = {'INSTRUME': 'MEGARA', 'INSCONF': 'ca3558e3-e50d-4bbc-86bd-da50a0998a48'}
    hdu = fits.PrimaryHDU(data=[[1]])
    for k, v in hdr.items():
        hdu.header[k] = v
    hdulist = fits.HDUList([hdu])
    frame = numina.core.DataFrame(frame=hdulist)
    return frame


@pytest.mark.parametrize("conf, uuix", [
    ['default', "ca3558e3-e50d-4bbc-86bd-da50a0998a48"],
    ["ca3558e3-e50d-4bbc-86bd-da50a0998a48", "ca3558e3-e50d-4bbc-86bd-da50a0998a48"],
    ["9a86b2b2-3f7d-48ec-8f4f-3780ec967c90", "9a86b2b2-3f7d-48ec-8f4f-3780ec967c90"],
    ["66f2283e-3049-4d4b-8ef1-14d62fcb611d", "66f2283e-3049-4d4b-8ef1-14d62fcb611d"],
    ["4fd05b24-2ed9-457b-b563-a3c618bb1d4c", "4fd05b24-2ed9-457b-b563-a3c618bb1d4c"]
])
def test_loader1(conf, uuix):
    import numina.core
    from numina.instrument.assembly import assembly_instrument

    obs = numina.core.ObservationResult(instrument='MEGARA')
    obs.frames.append(create_simple_frame())
    drpm = load_drp()
    obs.configuration = conf

    key, date_obs, keyname = drpm.select_profile(obs)
    ins = assembly_instrument(drpm.configurations, key, date_obs, by_key=keyname)
    assert isinstance(ins, numina.instrument.generic.InstrumentGeneric)
    assert str(ins.origin.uuid) == uuix




import numpy as np
from ..hexgrid import calc_matrix, M_SQRT3, calc_lcb_grid


def test_calc_matrix():
    res = calc_matrix(4, 4, grid_type=2)
    r1 = M_SQRT3 * np.array(
        [0, 0, 0, 0, 1.0/2, 1.0/2, 1.0/2, 1.0/2, 1, 1, 1, 1, 3.0/2, 3.0/2, 3.0/2, 3.0/2]
    )
    r2 = [0, 1, 2, 3, -0.5, 0.5, 1.5, 2.5, 0, 1, 2, 3, -0.5, 0.5, 1.5, 2.5]
    expected = np.array([r1, r2])
    assert np.allclose(res, expected)


def test_calc_lcb_grid():
    res_x, res_y = calc_lcb_grid()
    assert res_x.shape == (567,)
    assert res_y.shape == (567,)
    assert np.allclose([res_x.min(), res_x.max()], [-11.258330249197702, 11.258330249197702])
    assert np.allclose([res_y.min(), res_y.max()], [-10.5, 10.0])




import math

import astropy.wcs
import astropy.io.fits as fits


def create_simple_hdul():
    """Create a simple image for testing"""
    prim = fits.PrimaryHDU()
    prim.header['instrume'] = 'MEGARA'
    prim.header['VPH'] = 'LR-B'
    prim.header['DATE-OBS'] = '2019-02-21T01:02:02.2'
    prim.header['insmode'] = 'LCB'
    prim.header['UUID'] = '410f6ea0-c3df-481c-9820-5cf9e0ed3d91'
    fibers = fits.ImageHDU(name='FIBERS')
    fibers.header['CONFID'] = 'a908e9d2-7599-41f3-9e9d-91042df01da5'
    simple_img = fits.HDUList([prim, fibers])
    return simple_img


def generate_sky_wcs():
    wcsl = astropy.wcs.WCS(naxis=2)

    wcsl.wcs.crpix = [512, 512]
    wcsl.wcs.crval = [9.0000, 32.0000]
    wcsl.wcs.cdelt = [0.01, 0.01]
    wcsl.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    ang = math.pi / 3.0
    wcsl.wcs.pc = [[math.cos(ang), -math.sin(ang)], [math.sin(ang), math.cos(ang)]]
    return wcsl


def create_sky_header():
    wcsl = generate_sky_wcs()
    return wcsl.to_header()


def create_sky_header2():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2017-08-23T21:38:30.55'
    # The values of CRPIX and CDELT
    # are not in pixels
    # CRPIX is in mm off the center of the focal plane
    # and CDELT in deg / mm
    hdr['WCSAXES'] = 2
    hdr['CTYPE1'] = 'RA---TAN'
    hdr['CRPIX1'] = 0.0
    hdr['CRVAL1'] = 255.876802773371
    hdr['CDELT1'] = -0.000336666666666667
    hdr['CUNIT1'] = 'deg'
    hdr['CRPIX2'] = 0
    hdr['CRVAL2'] = 45.6801440902947
    hdr['CDELT2'] = 0.000336666666666667
    hdr['CTYPE2'] = 'DEC--TAN'
    hdr['CUNIT2'] = 'deg'
    hdr['PC1_1'] = 0.0104368794387827
    hdr['PC1_2'] = 0.999945534290533
    hdr['PC2_1'] = -0.999945534290533
    hdr['PC2_2'] = 0.0104368794387827
    hdr['LONPOLE'] = 180.0
    hdr['LATPOLE'] = hdr['CRVAL2']
    hdr['RADESYS'] = 'FK5'
    hdr['EQUINOX'] = 2000.
    return hdr

def create_spec_header():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2017-08-23T21:38:30.55'
    # GTC
    hdr['OBSGEO-X'] = 5327285.0921
    hdr['OBSGEO-Y'] = -1718777.1125
    hdr['OBSGEO-Z'] = 3051786.7327

    hdr['RADEG'] = 285.481037748898
    hdr['DECDEG'] = 42.4882140636786

    hdr['CTYPE1'] = 'AWAV'
    hdr['CRPIX1'] = 1
    hdr['CRVAL1'] = 362.0
    hdr['CDELT1'] = 1.86
    hdr['CUNIT1'] = 'nm'

    hdr['CRPIX2'] = 0
    hdr['CRVAL2'] = 0
    hdr['CDELT2'] = 1
    hdr['CTYPE2'] = ''
    return hdr


def create_spec_header2():
    hdr = fits.Header()
    hdr['DATE-OBS'] = '2020-06-24T02:53:22.03'
    # GTC
    hdr['OBSGEO-X'] = 5327285.0921
    hdr['OBSGEO-Y'] = -1718777.1125
    hdr['OBSGEO-Z'] = 3051786.7327
    hdr['OBSGEO-L'] = -17.881600
    hdr['OBSGEO-B'] = 28.760600
    hdr['OBSGEO-H'] = 2322.994

    hdr['RADEG'] = 255.876802773371
    hdr['DECDEG'] = 45.6801440902947

    hdr['CTYPE1'] = 'AWAV'
    hdr['CRPIX1'] = 1
    hdr['CRVAL1'] = 6030
    hdr['CDELT1'] = 0.31
    hdr['CUNIT1'] = 'Angstrom'

    hdr['CRPIX2'] = 0
    hdr['CRVAL2'] = 0
    hdr['CDELT2'] = 1
    hdr['CTYPE2'] = ''
    return hdr